import argparse
import asyncio
from solana.keypair import Keypair
from base58 import b58decode

from ..mango_client import MangoClient


async def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--symbol',
        required=True
    )

    parser.add_argument(
        '--mango-account',
        help='Mango account primary key.',
        required=True
    )

    parser.add_argument(
        '--keypair',
        help='Solana wallet private key, to sign transactions for the Mango account.',
        required=True
    )

    args = parser.parse_args()

    mango_client = await MangoClient.connect()

    mango_account = await mango_client.get_mango_account(args.mango_account)

    keypair = Keypair.from_secret_key(b58decode(args.keypair))

    print(await mango_client.cancel_all_orders(mango_account, keypair, args.symbol))

    # 5B9Rq1sZAjtzXFbYh7wiV1thNGiaAC716ejPSBLE3EvTNgKC8CmFETdtMx8L5nfXHYZ3R8WyAqr9upfRbVYyGVg5
    # (Check the UI)


if __name__ == '__main__':
    asyncio.run(main())



from mango_explorer_v4.types.health_type import HealthTypeKind
from mango_explorer_v4.types.token_info import TokenInfo
from mango_explorer_v4.types.i80f48 import I80F48
from mango_explorer_v4.types.serum3_info import Serum3Info
from mango_explorer_v4.helpers.token_info import TokenInfoHelper
from mango_explorer_v4.helpers.prices import PricesHelper
from mango_explorer_v4.constructs.serum3_reserved import Serum3Reserved
from decimal import Decimal

class Serum3InfoHelper:
    @staticmethod
    def health_contribution(
        serum3_info: Serum3Info,
        health_type: HealthTypeKind,
        token_infos: [TokenInfo],
        token_max_reserved: [I80F48],
        market_reserved: Serum3Reserved
    ) -> I80F48:
        if any([
            market_reserved.all_reserved_as_base == 0,
            market_reserved.all_reserved_as_quote == 0
        ]):
            return I80F48.from_decimal(Decimal(0))

        base_info, quote_info = token_infos[serum3_info.base_index], token_infos[serum3_info.quote_index]

        base_max_reserved, quote_max_reserved = token_max_reserved[serum3_info.base_index], token_max_reserved[serum3_info.quote_index]

        # How much would the health increase if the reserved balance were applied to the passed token info?
        def get_health_effect(token_info: TokenInfo, token_max_reserved: I80F48, market_reserved: I80F48) -> I80F48:
            # This balance includes all possible reserved funds from markets that relate to the
            # token, including this market itself: `tokenMaxReserved` is already included in `maxBalance`.
            max_balance = token_info.balance_native.to_decimal() + token_max_reserved.to_decimal()

            # Assuming `reserved` was added to `max_balance` last (because that gives the smallest
            # health effects): how much did health change because of it?

            if max_balance > market_reserved.to_decimal():
                asset_part, liab_part = market_reserved, 0
            elif max_balance < 0:
                asset_part, liab_part = 0, market_reserved
            else:
                asset_part, liab_part = max_balance, market_reserved.to_decimal() - max_balance

            if health_type is None:
                return I80F48.from_decimal(asset_part * token_info.prices.oracle.to_decimal() + liab_part * token_info.prices.oracle.to_decimal())

            asset_weight, liab_weight, asset_price, liab_price = [
                TokenInfoHelper.asset_weight(token_info, health_type),
                TokenInfoHelper.liab_weight(token_info, health_type),
                PricesHelper.asset(token_info.prices, health_type),
                PricesHelper.liab(token_info.prices, health_type)
            ]

            return I80F48.from_decimal(asset_weight * asset_part * asset_price + liab_weight * liab_part * liab_price)

        health_base = get_health_effect(base_info, base_max_reserved, I80F48.from_decimal(market_reserved.all_reserved_as_base))

        health_quote = get_health_effect(quote_info, quote_max_reserved, I80F48.from_decimal(market_reserved.all_reserved_as_quote))

        return I80F48.from_decimal(min(health_base.to_decimal(), health_quote.to_decimal()))





from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class GroupEditArgs(typing.TypedDict):
    admin_opt: typing.Optional[PublicKey]
    fast_listing_admin_opt: typing.Optional[PublicKey]
    security_admin_opt: typing.Optional[PublicKey]
    testing_opt: typing.Optional[int]
    version_opt: typing.Optional[int]
    deposit_limit_quote_opt: typing.Optional[int]
    buyback_fees_opt: typing.Optional[bool]
    buyback_fees_bonus_factor_opt: typing.Optional[float]
    buyback_fees_swap_mango_account_opt: typing.Optional[PublicKey]
    mngo_token_index_opt: typing.Optional[int]
    buyback_fees_expiry_interval_opt: typing.Optional[int]


layout = borsh.CStruct(
    "admin_opt" / borsh.Option(BorshPubkey),
    "fast_listing_admin_opt" / borsh.Option(BorshPubkey),
    "security_admin_opt" / borsh.Option(BorshPubkey),
    "testing_opt" / borsh.Option(borsh.U8),
    "version_opt" / borsh.Option(borsh.U8),
    "deposit_limit_quote_opt" / borsh.Option(borsh.U64),
    "buyback_fees_opt" / borsh.Option(borsh.Bool),
    "buyback_fees_bonus_factor_opt" / borsh.Option(borsh.F32),
    "buyback_fees_swap_mango_account_opt" / borsh.Option(BorshPubkey),
    "mngo_token_index_opt" / borsh.Option(borsh.U16),
    "buyback_fees_expiry_interval_opt" / borsh.Option(borsh.U64),
)


class GroupEditAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey


def group_edit(
    args: GroupEditArgs,
    accounts: GroupEditAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x08X\xb7\xf9\xa6s7\xe3"
    encoded_args = layout.build(
        {
            "admin_opt": args["admin_opt"],
            "fast_listing_admin_opt": args["fast_listing_admin_opt"],
            "security_admin_opt": args["security_admin_opt"],
            "testing_opt": args["testing_opt"],
            "version_opt": args["version_opt"],
            "deposit_limit_quote_opt": args["deposit_limit_quote_opt"],
            "buyback_fees_opt": args["buyback_fees_opt"],
            "buyback_fees_bonus_factor_opt": args["buyback_fees_bonus_factor_opt"],
            "buyback_fees_swap_mango_account_opt": args[
                "buyback_fees_swap_mango_account_opt"
            ],
            "mngo_token_index_opt": args["mngo_token_index_opt"],
            "buyback_fees_expiry_interval_opt": args[
                "buyback_fees_expiry_interval_opt"
            ],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



import typing
from dataclasses import dataclass
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class MintInfoJSON(typing.TypedDict):
    group: str
    token_index: int
    group_insurance_fund: int
    padding1: list[int]
    mint: str
    banks: list[str]
    vaults: list[str]
    oracle: str
    registration_time: int
    reserved: list[int]


@dataclass
class MintInfo:
    discriminator: typing.ClassVar = b"\xc7s\xd5\xdd\xdb\x1d\x87\xae"
    layout: typing.ClassVar = borsh.CStruct(
        "group" / BorshPubkey,
        "token_index" / borsh.U16,
        "group_insurance_fund" / borsh.U8,
        "padding1" / borsh.U8[5],
        "mint" / BorshPubkey,
        "banks" / BorshPubkey[6],
        "vaults" / BorshPubkey[6],
        "oracle" / BorshPubkey,
        "registration_time" / borsh.U64,
        "reserved" / borsh.U8[2560],
    )
    group: PublicKey
    token_index: int
    group_insurance_fund: int
    padding1: list[int]
    mint: PublicKey
    banks: list[PublicKey]
    vaults: list[PublicKey]
    oracle: PublicKey
    registration_time: int
    reserved: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.Optional["MintInfo"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["MintInfo"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["MintInfo"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "MintInfo":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = MintInfo.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            group=dec.group,
            token_index=dec.token_index,
            group_insurance_fund=dec.group_insurance_fund,
            padding1=dec.padding1,
            mint=dec.mint,
            banks=dec.banks,
            vaults=dec.vaults,
            oracle=dec.oracle,
            registration_time=dec.registration_time,
            reserved=dec.reserved,
        )

    def to_json(self) -> MintInfoJSON:
        return {
            "group": str(self.group),
            "token_index": self.token_index,
            "group_insurance_fund": self.group_insurance_fund,
            "padding1": self.padding1,
            "mint": str(self.mint),
            "banks": list(map(lambda item: str(item), self.banks)),
            "vaults": list(map(lambda item: str(item), self.vaults)),
            "oracle": str(self.oracle),
            "registration_time": self.registration_time,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: MintInfoJSON) -> "MintInfo":
        return cls(
            group=PublicKey(obj["group"]),
            token_index=obj["token_index"],
            group_insurance_fund=obj["group_insurance_fund"],
            padding1=obj["padding1"],
            mint=PublicKey(obj["mint"]),
            banks=list(map(lambda item: PublicKey(item), obj["banks"])),
            vaults=list(map(lambda item: PublicKey(item), obj["vaults"])),
            oracle=PublicKey(obj["oracle"]),
            registration_time=obj["registration_time"],
            reserved=obj["reserved"],
        )



import typing
from decimal import Decimal

import construct
from solana.publickey import PublicKey


class PublicKeyAdapter(construct.Adapter):
    def __init__(self) -> None:
        super().__init__(construct.Bytes(32))

    def _decode(self, obj: bytes, context: typing.Any, path: typing.Any) -> typing.Optional[PublicKey]:
        if (obj is None) or (obj == bytes([0] * 32)):
            return None
        return PublicKey(obj)

    def _encode(self, obj: PublicKey, context: typing.Any, path: typing.Any) -> bytes:
        return bytes(obj)


class DecimalAdapter(construct.Adapter):
    def __init__(self, size: int = 8) -> None:
        super().__init__(construct.BytesInteger(size, swapped=True))

    def _decode(self, obj: int, context: typing.Any, path: typing.Any) -> Decimal:
        return Decimal(obj)

    def _encode(self, obj: Decimal, context: typing.Any, path: typing.Any) -> int:
        # Can only encode Decimal values.
        return int(obj)


PRICE_INFO = construct.Struct(
    "price" / DecimalAdapter(),
    "conf" / DecimalAdapter(),
    "status" / construct.Enum(construct.Int32ul, Unknown=0, Trading=1, Halted=2, Auction=3),
    "corp_act" / construct.Enum(construct.Int32ul, NoCorpAct=0),
    "pub_slot" / DecimalAdapter(),
)

ACCOUNT_TYPE = construct.Enum(
    construct.Int32ul, Unknown=0, Mapping=1, Product=2, Price=3
)

PRICE_COMP = construct.Struct(
    "publisher" / PublicKeyAdapter(), "agg" / PRICE_INFO, "latest" / PRICE_INFO
)

PRICE = construct.Struct(
    "magic" / DecimalAdapter(4),
    "ver" / DecimalAdapter(4),
    "atype" / ACCOUNT_TYPE,
    "size" / DecimalAdapter(4),
    "ptype" / construct.Enum(construct.Int32ul, Unknown=0, Price=1),
    "expo" / construct.Int32sl,
    "num" / DecimalAdapter(4),
    "unused" / DecimalAdapter(4),
    "curr_slot" / DecimalAdapter(),
    "valid_slot" / DecimalAdapter(),
    "drv" / construct.Array(8, construct.Int64sl),
    "prod" / PublicKeyAdapter(),
    "next" / PublicKeyAdapter(),
    "agg_pub" / PublicKeyAdapter(),
    "agg" / PRICE_INFO,
    "comp" / construct.Array(32, PRICE_COMP),
)



from __future__ import annotations
from . import (
    post_order_type,
)
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class ImmediateOrCancelJSONValue(typing.TypedDict):
    price_lots: int


class FixedJSONValue(typing.TypedDict):
    price_lots: int
    order_type: post_order_type.PostOrderTypeJSON


class OraclePeggedJSONValue(typing.TypedDict):
    price_offset_lots: int
    order_type: post_order_type.PostOrderTypeJSON
    peg_limit: int
    max_oracle_staleness_slots: int


class ImmediateOrCancelValue(typing.TypedDict):
    price_lots: int


class FixedValue(typing.TypedDict):
    price_lots: int
    order_type: post_order_type.PostOrderTypeKind


class OraclePeggedValue(typing.TypedDict):
    price_offset_lots: int
    order_type: post_order_type.PostOrderTypeKind
    peg_limit: int
    max_oracle_staleness_slots: int


class MarketJSON(typing.TypedDict):
    kind: typing.Literal["Market"]


class ImmediateOrCancelJSON(typing.TypedDict):
    value: ImmediateOrCancelJSONValue
    kind: typing.Literal["ImmediateOrCancel"]


class FixedJSON(typing.TypedDict):
    value: FixedJSONValue
    kind: typing.Literal["Fixed"]


class OraclePeggedJSON(typing.TypedDict):
    value: OraclePeggedJSONValue
    kind: typing.Literal["OraclePegged"]


@dataclass
class Market:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Market"

    @classmethod
    def to_json(cls) -> MarketJSON:
        return MarketJSON(
            kind="Market",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Market": {},
        }


@dataclass
class ImmediateOrCancel:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "ImmediateOrCancel"
    value: ImmediateOrCancelValue

    def to_json(self) -> ImmediateOrCancelJSON:
        return ImmediateOrCancelJSON(
            kind="ImmediateOrCancel",
            value={
                "price_lots": self.value["price_lots"],
            },
        )

    def to_encodable(self) -> dict:
        return {
            "ImmediateOrCancel": {
                "price_lots": self.value["price_lots"],
            },
        }


@dataclass
class Fixed:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Fixed"
    value: FixedValue

    def to_json(self) -> FixedJSON:
        return FixedJSON(
            kind="Fixed",
            value={
                "price_lots": self.value["price_lots"],
                "order_type": self.value["order_type"].to_json(),
            },
        )

    def to_encodable(self) -> dict:
        return {
            "Fixed": {
                "price_lots": self.value["price_lots"],
                "order_type": self.value["order_type"].to_encodable(),
            },
        }


@dataclass
class OraclePegged:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "OraclePegged"
    value: OraclePeggedValue

    def to_json(self) -> OraclePeggedJSON:
        return OraclePeggedJSON(
            kind="OraclePegged",
            value={
                "price_offset_lots": self.value["price_offset_lots"],
                "order_type": self.value["order_type"].to_json(),
                "peg_limit": self.value["peg_limit"],
                "max_oracle_staleness_slots": self.value["max_oracle_staleness_slots"],
            },
        )

    def to_encodable(self) -> dict:
        return {
            "OraclePegged": {
                "price_offset_lots": self.value["price_offset_lots"],
                "order_type": self.value["order_type"].to_encodable(),
                "peg_limit": self.value["peg_limit"],
                "max_oracle_staleness_slots": self.value["max_oracle_staleness_slots"],
            },
        }


OrderParamsKind = typing.Union[Market, ImmediateOrCancel, Fixed, OraclePegged]
OrderParamsJSON = typing.Union[
    MarketJSON, ImmediateOrCancelJSON, FixedJSON, OraclePeggedJSON
]


def from_decoded(obj: dict) -> OrderParamsKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Market" in obj:
        return Market()
    if "ImmediateOrCancel" in obj:
        val = obj["ImmediateOrCancel"]
        return ImmediateOrCancel(
            ImmediateOrCancelValue(
                price_lots=val["price_lots"],
            )
        )
    if "Fixed" in obj:
        val = obj["Fixed"]
        return Fixed(
            FixedValue(
                price_lots=val["price_lots"],
                order_type=post_order_type.from_decoded(val["order_type"]),
            )
        )
    if "OraclePegged" in obj:
        val = obj["OraclePegged"]
        return OraclePegged(
            OraclePeggedValue(
                price_offset_lots=val["price_offset_lots"],
                order_type=post_order_type.from_decoded(val["order_type"]),
                peg_limit=val["peg_limit"],
                max_oracle_staleness_slots=val["max_oracle_staleness_slots"],
            )
        )
    raise ValueError("Invalid enum object")


def from_json(obj: OrderParamsJSON) -> OrderParamsKind:
    if obj["kind"] == "Market":
        return Market()
    if obj["kind"] == "ImmediateOrCancel":
        immediate_or_cancel_json_value = typing.cast(
            ImmediateOrCancelJSONValue, obj["value"]
        )
        return ImmediateOrCancel(
            ImmediateOrCancelValue(
                price_lots=immediate_or_cancel_json_value["price_lots"],
            )
        )
    if obj["kind"] == "Fixed":
        fixed_json_value = typing.cast(FixedJSONValue, obj["value"])
        return Fixed(
            FixedValue(
                price_lots=fixed_json_value["price_lots"],
                order_type=post_order_type.from_json(fixed_json_value["order_type"]),
            )
        )
    if obj["kind"] == "OraclePegged":
        oracle_pegged_json_value = typing.cast(OraclePeggedJSONValue, obj["value"])
        return OraclePegged(
            OraclePeggedValue(
                price_offset_lots=oracle_pegged_json_value["price_offset_lots"],
                order_type=post_order_type.from_json(
                    oracle_pegged_json_value["order_type"]
                ),
                peg_limit=oracle_pegged_json_value["peg_limit"],
                max_oracle_staleness_slots=oracle_pegged_json_value[
                    "max_oracle_staleness_slots"
                ],
            )
        )
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Market" / borsh.CStruct(),
    "ImmediateOrCancel" / borsh.CStruct("price_lots" / borsh.I64),
    "Fixed"
    / borsh.CStruct("price_lots" / borsh.I64, "order_type" / post_order_type.layout),
    "OraclePegged"
    / borsh.CStruct(
        "price_offset_lots" / borsh.I64,
        "order_type" / post_order_type.layout,
        "peg_limit" / borsh.I64,
        "max_oracle_staleness_slots" / borsh.I32,
    ),
)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class AccountCloseJSON(typing.TypedDict):
    kind: typing.Literal["AccountClose"]


class AccountCreateJSON(typing.TypedDict):
    kind: typing.Literal["AccountCreate"]


class AccountEditJSON(typing.TypedDict):
    kind: typing.Literal["AccountEdit"]


class AccountExpandJSON(typing.TypedDict):
    kind: typing.Literal["AccountExpand"]


class AccountToggleFreezeJSON(typing.TypedDict):
    kind: typing.Literal["AccountToggleFreeze"]


class AltExtendJSON(typing.TypedDict):
    kind: typing.Literal["AltExtend"]


class AltSetJSON(typing.TypedDict):
    kind: typing.Literal["AltSet"]


class FlashLoanJSON(typing.TypedDict):
    kind: typing.Literal["FlashLoan"]


class GroupCloseJSON(typing.TypedDict):
    kind: typing.Literal["GroupClose"]


class GroupCreateJSON(typing.TypedDict):
    kind: typing.Literal["GroupCreate"]


class HealthRegionJSON(typing.TypedDict):
    kind: typing.Literal["HealthRegion"]


class PerpCancelAllOrdersJSON(typing.TypedDict):
    kind: typing.Literal["PerpCancelAllOrders"]


class PerpCancelAllOrdersBySideJSON(typing.TypedDict):
    kind: typing.Literal["PerpCancelAllOrdersBySide"]


class PerpCancelOrderJSON(typing.TypedDict):
    kind: typing.Literal["PerpCancelOrder"]


class PerpCancelOrderByClientOrderIdJSON(typing.TypedDict):
    kind: typing.Literal["PerpCancelOrderByClientOrderId"]


class PerpCloseMarketJSON(typing.TypedDict):
    kind: typing.Literal["PerpCloseMarket"]


class PerpConsumeEventsJSON(typing.TypedDict):
    kind: typing.Literal["PerpConsumeEvents"]


class PerpCreateMarketJSON(typing.TypedDict):
    kind: typing.Literal["PerpCreateMarket"]


class PerpDeactivatePositionJSON(typing.TypedDict):
    kind: typing.Literal["PerpDeactivatePosition"]


class PerpLiqBaseOrPositivePnlJSON(typing.TypedDict):
    kind: typing.Literal["PerpLiqBaseOrPositivePnl"]


class PerpLiqForceCancelOrdersJSON(typing.TypedDict):
    kind: typing.Literal["PerpLiqForceCancelOrders"]


class PerpLiqNegativePnlOrBankruptcyJSON(typing.TypedDict):
    kind: typing.Literal["PerpLiqNegativePnlOrBankruptcy"]


class PerpPlaceOrderJSON(typing.TypedDict):
    kind: typing.Literal["PerpPlaceOrder"]


class PerpSettleFeesJSON(typing.TypedDict):
    kind: typing.Literal["PerpSettleFees"]


class PerpSettlePnlJSON(typing.TypedDict):
    kind: typing.Literal["PerpSettlePnl"]


class PerpUpdateFundingJSON(typing.TypedDict):
    kind: typing.Literal["PerpUpdateFunding"]


class Serum3CancelAllOrdersJSON(typing.TypedDict):
    kind: typing.Literal["Serum3CancelAllOrders"]


class Serum3CancelOrderJSON(typing.TypedDict):
    kind: typing.Literal["Serum3CancelOrder"]


class Serum3CloseOpenOrdersJSON(typing.TypedDict):
    kind: typing.Literal["Serum3CloseOpenOrders"]


class Serum3CreateOpenOrdersJSON(typing.TypedDict):
    kind: typing.Literal["Serum3CreateOpenOrders"]


class Serum3DeregisterMarketJSON(typing.TypedDict):
    kind: typing.Literal["Serum3DeregisterMarket"]


class Serum3EditMarketJSON(typing.TypedDict):
    kind: typing.Literal["Serum3EditMarket"]


class Serum3LiqForceCancelOrdersJSON(typing.TypedDict):
    kind: typing.Literal["Serum3LiqForceCancelOrders"]


class Serum3PlaceOrderJSON(typing.TypedDict):
    kind: typing.Literal["Serum3PlaceOrder"]


class Serum3RegisterMarketJSON(typing.TypedDict):
    kind: typing.Literal["Serum3RegisterMarket"]


class Serum3SettleFundsJSON(typing.TypedDict):
    kind: typing.Literal["Serum3SettleFunds"]


class StubOracleCloseJSON(typing.TypedDict):
    kind: typing.Literal["StubOracleClose"]


class StubOracleCreateJSON(typing.TypedDict):
    kind: typing.Literal["StubOracleCreate"]


class StubOracleSetJSON(typing.TypedDict):
    kind: typing.Literal["StubOracleSet"]


class TokenAddBankJSON(typing.TypedDict):
    kind: typing.Literal["TokenAddBank"]


class TokenDepositJSON(typing.TypedDict):
    kind: typing.Literal["TokenDeposit"]


class TokenDeregisterJSON(typing.TypedDict):
    kind: typing.Literal["TokenDeregister"]


class TokenLiqBankruptcyJSON(typing.TypedDict):
    kind: typing.Literal["TokenLiqBankruptcy"]


class TokenLiqWithTokenJSON(typing.TypedDict):
    kind: typing.Literal["TokenLiqWithToken"]


class TokenRegisterJSON(typing.TypedDict):
    kind: typing.Literal["TokenRegister"]


class TokenRegisterTrustlessJSON(typing.TypedDict):
    kind: typing.Literal["TokenRegisterTrustless"]


class TokenUpdateIndexAndRateJSON(typing.TypedDict):
    kind: typing.Literal["TokenUpdateIndexAndRate"]


class TokenWithdrawJSON(typing.TypedDict):
    kind: typing.Literal["TokenWithdraw"]


class AccountBuybackFeesWithMngoJSON(typing.TypedDict):
    kind: typing.Literal["AccountBuybackFeesWithMngo"]


@dataclass
class AccountClose:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "AccountClose"

    @classmethod
    def to_json(cls) -> AccountCloseJSON:
        return AccountCloseJSON(
            kind="AccountClose",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AccountClose": {},
        }


@dataclass
class AccountCreate:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "AccountCreate"

    @classmethod
    def to_json(cls) -> AccountCreateJSON:
        return AccountCreateJSON(
            kind="AccountCreate",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AccountCreate": {},
        }


@dataclass
class AccountEdit:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "AccountEdit"

    @classmethod
    def to_json(cls) -> AccountEditJSON:
        return AccountEditJSON(
            kind="AccountEdit",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AccountEdit": {},
        }


@dataclass
class AccountExpand:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "AccountExpand"

    @classmethod
    def to_json(cls) -> AccountExpandJSON:
        return AccountExpandJSON(
            kind="AccountExpand",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AccountExpand": {},
        }


@dataclass
class AccountToggleFreeze:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "AccountToggleFreeze"

    @classmethod
    def to_json(cls) -> AccountToggleFreezeJSON:
        return AccountToggleFreezeJSON(
            kind="AccountToggleFreeze",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AccountToggleFreeze": {},
        }


@dataclass
class AltExtend:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "AltExtend"

    @classmethod
    def to_json(cls) -> AltExtendJSON:
        return AltExtendJSON(
            kind="AltExtend",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AltExtend": {},
        }


@dataclass
class AltSet:
    discriminator: typing.ClassVar = 6
    kind: typing.ClassVar = "AltSet"

    @classmethod
    def to_json(cls) -> AltSetJSON:
        return AltSetJSON(
            kind="AltSet",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AltSet": {},
        }


@dataclass
class FlashLoan:
    discriminator: typing.ClassVar = 7
    kind: typing.ClassVar = "FlashLoan"

    @classmethod
    def to_json(cls) -> FlashLoanJSON:
        return FlashLoanJSON(
            kind="FlashLoan",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "FlashLoan": {},
        }


@dataclass
class GroupClose:
    discriminator: typing.ClassVar = 8
    kind: typing.ClassVar = "GroupClose"

    @classmethod
    def to_json(cls) -> GroupCloseJSON:
        return GroupCloseJSON(
            kind="GroupClose",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "GroupClose": {},
        }


@dataclass
class GroupCreate:
    discriminator: typing.ClassVar = 9
    kind: typing.ClassVar = "GroupCreate"

    @classmethod
    def to_json(cls) -> GroupCreateJSON:
        return GroupCreateJSON(
            kind="GroupCreate",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "GroupCreate": {},
        }


@dataclass
class HealthRegion:
    discriminator: typing.ClassVar = 10
    kind: typing.ClassVar = "HealthRegion"

    @classmethod
    def to_json(cls) -> HealthRegionJSON:
        return HealthRegionJSON(
            kind="HealthRegion",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "HealthRegion": {},
        }


@dataclass
class PerpCancelAllOrders:
    discriminator: typing.ClassVar = 11
    kind: typing.ClassVar = "PerpCancelAllOrders"

    @classmethod
    def to_json(cls) -> PerpCancelAllOrdersJSON:
        return PerpCancelAllOrdersJSON(
            kind="PerpCancelAllOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpCancelAllOrders": {},
        }


@dataclass
class PerpCancelAllOrdersBySide:
    discriminator: typing.ClassVar = 12
    kind: typing.ClassVar = "PerpCancelAllOrdersBySide"

    @classmethod
    def to_json(cls) -> PerpCancelAllOrdersBySideJSON:
        return PerpCancelAllOrdersBySideJSON(
            kind="PerpCancelAllOrdersBySide",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpCancelAllOrdersBySide": {},
        }


@dataclass
class PerpCancelOrder:
    discriminator: typing.ClassVar = 13
    kind: typing.ClassVar = "PerpCancelOrder"

    @classmethod
    def to_json(cls) -> PerpCancelOrderJSON:
        return PerpCancelOrderJSON(
            kind="PerpCancelOrder",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpCancelOrder": {},
        }


@dataclass
class PerpCancelOrderByClientOrderId:
    discriminator: typing.ClassVar = 14
    kind: typing.ClassVar = "PerpCancelOrderByClientOrderId"

    @classmethod
    def to_json(cls) -> PerpCancelOrderByClientOrderIdJSON:
        return PerpCancelOrderByClientOrderIdJSON(
            kind="PerpCancelOrderByClientOrderId",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpCancelOrderByClientOrderId": {},
        }


@dataclass
class PerpCloseMarket:
    discriminator: typing.ClassVar = 15
    kind: typing.ClassVar = "PerpCloseMarket"

    @classmethod
    def to_json(cls) -> PerpCloseMarketJSON:
        return PerpCloseMarketJSON(
            kind="PerpCloseMarket",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpCloseMarket": {},
        }


@dataclass
class PerpConsumeEvents:
    discriminator: typing.ClassVar = 16
    kind: typing.ClassVar = "PerpConsumeEvents"

    @classmethod
    def to_json(cls) -> PerpConsumeEventsJSON:
        return PerpConsumeEventsJSON(
            kind="PerpConsumeEvents",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpConsumeEvents": {},
        }


@dataclass
class PerpCreateMarket:
    discriminator: typing.ClassVar = 17
    kind: typing.ClassVar = "PerpCreateMarket"

    @classmethod
    def to_json(cls) -> PerpCreateMarketJSON:
        return PerpCreateMarketJSON(
            kind="PerpCreateMarket",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpCreateMarket": {},
        }


@dataclass
class PerpDeactivatePosition:
    discriminator: typing.ClassVar = 18
    kind: typing.ClassVar = "PerpDeactivatePosition"

    @classmethod
    def to_json(cls) -> PerpDeactivatePositionJSON:
        return PerpDeactivatePositionJSON(
            kind="PerpDeactivatePosition",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpDeactivatePosition": {},
        }


@dataclass
class PerpLiqBaseOrPositivePnl:
    discriminator: typing.ClassVar = 19
    kind: typing.ClassVar = "PerpLiqBaseOrPositivePnl"

    @classmethod
    def to_json(cls) -> PerpLiqBaseOrPositivePnlJSON:
        return PerpLiqBaseOrPositivePnlJSON(
            kind="PerpLiqBaseOrPositivePnl",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpLiqBaseOrPositivePnl": {},
        }


@dataclass
class PerpLiqForceCancelOrders:
    discriminator: typing.ClassVar = 20
    kind: typing.ClassVar = "PerpLiqForceCancelOrders"

    @classmethod
    def to_json(cls) -> PerpLiqForceCancelOrdersJSON:
        return PerpLiqForceCancelOrdersJSON(
            kind="PerpLiqForceCancelOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpLiqForceCancelOrders": {},
        }


@dataclass
class PerpLiqNegativePnlOrBankruptcy:
    discriminator: typing.ClassVar = 21
    kind: typing.ClassVar = "PerpLiqNegativePnlOrBankruptcy"

    @classmethod
    def to_json(cls) -> PerpLiqNegativePnlOrBankruptcyJSON:
        return PerpLiqNegativePnlOrBankruptcyJSON(
            kind="PerpLiqNegativePnlOrBankruptcy",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpLiqNegativePnlOrBankruptcy": {},
        }


@dataclass
class PerpPlaceOrder:
    discriminator: typing.ClassVar = 22
    kind: typing.ClassVar = "PerpPlaceOrder"

    @classmethod
    def to_json(cls) -> PerpPlaceOrderJSON:
        return PerpPlaceOrderJSON(
            kind="PerpPlaceOrder",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpPlaceOrder": {},
        }


@dataclass
class PerpSettleFees:
    discriminator: typing.ClassVar = 23
    kind: typing.ClassVar = "PerpSettleFees"

    @classmethod
    def to_json(cls) -> PerpSettleFeesJSON:
        return PerpSettleFeesJSON(
            kind="PerpSettleFees",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpSettleFees": {},
        }


@dataclass
class PerpSettlePnl:
    discriminator: typing.ClassVar = 24
    kind: typing.ClassVar = "PerpSettlePnl"

    @classmethod
    def to_json(cls) -> PerpSettlePnlJSON:
        return PerpSettlePnlJSON(
            kind="PerpSettlePnl",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpSettlePnl": {},
        }


@dataclass
class PerpUpdateFunding:
    discriminator: typing.ClassVar = 25
    kind: typing.ClassVar = "PerpUpdateFunding"

    @classmethod
    def to_json(cls) -> PerpUpdateFundingJSON:
        return PerpUpdateFundingJSON(
            kind="PerpUpdateFunding",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "PerpUpdateFunding": {},
        }


@dataclass
class Serum3CancelAllOrders:
    discriminator: typing.ClassVar = 26
    kind: typing.ClassVar = "Serum3CancelAllOrders"

    @classmethod
    def to_json(cls) -> Serum3CancelAllOrdersJSON:
        return Serum3CancelAllOrdersJSON(
            kind="Serum3CancelAllOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3CancelAllOrders": {},
        }


@dataclass
class Serum3CancelOrder:
    discriminator: typing.ClassVar = 27
    kind: typing.ClassVar = "Serum3CancelOrder"

    @classmethod
    def to_json(cls) -> Serum3CancelOrderJSON:
        return Serum3CancelOrderJSON(
            kind="Serum3CancelOrder",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3CancelOrder": {},
        }


@dataclass
class Serum3CloseOpenOrders:
    discriminator: typing.ClassVar = 28
    kind: typing.ClassVar = "Serum3CloseOpenOrders"

    @classmethod
    def to_json(cls) -> Serum3CloseOpenOrdersJSON:
        return Serum3CloseOpenOrdersJSON(
            kind="Serum3CloseOpenOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3CloseOpenOrders": {},
        }


@dataclass
class Serum3CreateOpenOrders:
    discriminator: typing.ClassVar = 29
    kind: typing.ClassVar = "Serum3CreateOpenOrders"

    @classmethod
    def to_json(cls) -> Serum3CreateOpenOrdersJSON:
        return Serum3CreateOpenOrdersJSON(
            kind="Serum3CreateOpenOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3CreateOpenOrders": {},
        }


@dataclass
class Serum3DeregisterMarket:
    discriminator: typing.ClassVar = 30
    kind: typing.ClassVar = "Serum3DeregisterMarket"

    @classmethod
    def to_json(cls) -> Serum3DeregisterMarketJSON:
        return Serum3DeregisterMarketJSON(
            kind="Serum3DeregisterMarket",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3DeregisterMarket": {},
        }


@dataclass
class Serum3EditMarket:
    discriminator: typing.ClassVar = 31
    kind: typing.ClassVar = "Serum3EditMarket"

    @classmethod
    def to_json(cls) -> Serum3EditMarketJSON:
        return Serum3EditMarketJSON(
            kind="Serum3EditMarket",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3EditMarket": {},
        }


@dataclass
class Serum3LiqForceCancelOrders:
    discriminator: typing.ClassVar = 32
    kind: typing.ClassVar = "Serum3LiqForceCancelOrders"

    @classmethod
    def to_json(cls) -> Serum3LiqForceCancelOrdersJSON:
        return Serum3LiqForceCancelOrdersJSON(
            kind="Serum3LiqForceCancelOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3LiqForceCancelOrders": {},
        }


@dataclass
class Serum3PlaceOrder:
    discriminator: typing.ClassVar = 33
    kind: typing.ClassVar = "Serum3PlaceOrder"

    @classmethod
    def to_json(cls) -> Serum3PlaceOrderJSON:
        return Serum3PlaceOrderJSON(
            kind="Serum3PlaceOrder",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3PlaceOrder": {},
        }


@dataclass
class Serum3RegisterMarket:
    discriminator: typing.ClassVar = 34
    kind: typing.ClassVar = "Serum3RegisterMarket"

    @classmethod
    def to_json(cls) -> Serum3RegisterMarketJSON:
        return Serum3RegisterMarketJSON(
            kind="Serum3RegisterMarket",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3RegisterMarket": {},
        }


@dataclass
class Serum3SettleFunds:
    discriminator: typing.ClassVar = 35
    kind: typing.ClassVar = "Serum3SettleFunds"

    @classmethod
    def to_json(cls) -> Serum3SettleFundsJSON:
        return Serum3SettleFundsJSON(
            kind="Serum3SettleFunds",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3SettleFunds": {},
        }


@dataclass
class StubOracleClose:
    discriminator: typing.ClassVar = 36
    kind: typing.ClassVar = "StubOracleClose"

    @classmethod
    def to_json(cls) -> StubOracleCloseJSON:
        return StubOracleCloseJSON(
            kind="StubOracleClose",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "StubOracleClose": {},
        }


@dataclass
class StubOracleCreate:
    discriminator: typing.ClassVar = 37
    kind: typing.ClassVar = "StubOracleCreate"

    @classmethod
    def to_json(cls) -> StubOracleCreateJSON:
        return StubOracleCreateJSON(
            kind="StubOracleCreate",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "StubOracleCreate": {},
        }


@dataclass
class StubOracleSet:
    discriminator: typing.ClassVar = 38
    kind: typing.ClassVar = "StubOracleSet"

    @classmethod
    def to_json(cls) -> StubOracleSetJSON:
        return StubOracleSetJSON(
            kind="StubOracleSet",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "StubOracleSet": {},
        }


@dataclass
class TokenAddBank:
    discriminator: typing.ClassVar = 39
    kind: typing.ClassVar = "TokenAddBank"

    @classmethod
    def to_json(cls) -> TokenAddBankJSON:
        return TokenAddBankJSON(
            kind="TokenAddBank",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenAddBank": {},
        }


@dataclass
class TokenDeposit:
    discriminator: typing.ClassVar = 40
    kind: typing.ClassVar = "TokenDeposit"

    @classmethod
    def to_json(cls) -> TokenDepositJSON:
        return TokenDepositJSON(
            kind="TokenDeposit",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenDeposit": {},
        }


@dataclass
class TokenDeregister:
    discriminator: typing.ClassVar = 41
    kind: typing.ClassVar = "TokenDeregister"

    @classmethod
    def to_json(cls) -> TokenDeregisterJSON:
        return TokenDeregisterJSON(
            kind="TokenDeregister",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenDeregister": {},
        }


@dataclass
class TokenLiqBankruptcy:
    discriminator: typing.ClassVar = 42
    kind: typing.ClassVar = "TokenLiqBankruptcy"

    @classmethod
    def to_json(cls) -> TokenLiqBankruptcyJSON:
        return TokenLiqBankruptcyJSON(
            kind="TokenLiqBankruptcy",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenLiqBankruptcy": {},
        }


@dataclass
class TokenLiqWithToken:
    discriminator: typing.ClassVar = 43
    kind: typing.ClassVar = "TokenLiqWithToken"

    @classmethod
    def to_json(cls) -> TokenLiqWithTokenJSON:
        return TokenLiqWithTokenJSON(
            kind="TokenLiqWithToken",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenLiqWithToken": {},
        }


@dataclass
class TokenRegister:
    discriminator: typing.ClassVar = 44
    kind: typing.ClassVar = "TokenRegister"

    @classmethod
    def to_json(cls) -> TokenRegisterJSON:
        return TokenRegisterJSON(
            kind="TokenRegister",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenRegister": {},
        }


@dataclass
class TokenRegisterTrustless:
    discriminator: typing.ClassVar = 45
    kind: typing.ClassVar = "TokenRegisterTrustless"

    @classmethod
    def to_json(cls) -> TokenRegisterTrustlessJSON:
        return TokenRegisterTrustlessJSON(
            kind="TokenRegisterTrustless",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenRegisterTrustless": {},
        }


@dataclass
class TokenUpdateIndexAndRate:
    discriminator: typing.ClassVar = 46
    kind: typing.ClassVar = "TokenUpdateIndexAndRate"

    @classmethod
    def to_json(cls) -> TokenUpdateIndexAndRateJSON:
        return TokenUpdateIndexAndRateJSON(
            kind="TokenUpdateIndexAndRate",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenUpdateIndexAndRate": {},
        }


@dataclass
class TokenWithdraw:
    discriminator: typing.ClassVar = 47
    kind: typing.ClassVar = "TokenWithdraw"

    @classmethod
    def to_json(cls) -> TokenWithdrawJSON:
        return TokenWithdrawJSON(
            kind="TokenWithdraw",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenWithdraw": {},
        }


@dataclass
class AccountBuybackFeesWithMngo:
    discriminator: typing.ClassVar = 48
    kind: typing.ClassVar = "AccountBuybackFeesWithMngo"

    @classmethod
    def to_json(cls) -> AccountBuybackFeesWithMngoJSON:
        return AccountBuybackFeesWithMngoJSON(
            kind="AccountBuybackFeesWithMngo",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AccountBuybackFeesWithMngo": {},
        }


IxGateKind = typing.Union[
    AccountClose,
    AccountCreate,
    AccountEdit,
    AccountExpand,
    AccountToggleFreeze,
    AltExtend,
    AltSet,
    FlashLoan,
    GroupClose,
    GroupCreate,
    HealthRegion,
    PerpCancelAllOrders,
    PerpCancelAllOrdersBySide,
    PerpCancelOrder,
    PerpCancelOrderByClientOrderId,
    PerpCloseMarket,
    PerpConsumeEvents,
    PerpCreateMarket,
    PerpDeactivatePosition,
    PerpLiqBaseOrPositivePnl,
    PerpLiqForceCancelOrders,
    PerpLiqNegativePnlOrBankruptcy,
    PerpPlaceOrder,
    PerpSettleFees,
    PerpSettlePnl,
    PerpUpdateFunding,
    Serum3CancelAllOrders,
    Serum3CancelOrder,
    Serum3CloseOpenOrders,
    Serum3CreateOpenOrders,
    Serum3DeregisterMarket,
    Serum3EditMarket,
    Serum3LiqForceCancelOrders,
    Serum3PlaceOrder,
    Serum3RegisterMarket,
    Serum3SettleFunds,
    StubOracleClose,
    StubOracleCreate,
    StubOracleSet,
    TokenAddBank,
    TokenDeposit,
    TokenDeregister,
    TokenLiqBankruptcy,
    TokenLiqWithToken,
    TokenRegister,
    TokenRegisterTrustless,
    TokenUpdateIndexAndRate,
    TokenWithdraw,
    AccountBuybackFeesWithMngo,
]
IxGateJSON = typing.Union[
    AccountCloseJSON,
    AccountCreateJSON,
    AccountEditJSON,
    AccountExpandJSON,
    AccountToggleFreezeJSON,
    AltExtendJSON,
    AltSetJSON,
    FlashLoanJSON,
    GroupCloseJSON,
    GroupCreateJSON,
    HealthRegionJSON,
    PerpCancelAllOrdersJSON,
    PerpCancelAllOrdersBySideJSON,
    PerpCancelOrderJSON,
    PerpCancelOrderByClientOrderIdJSON,
    PerpCloseMarketJSON,
    PerpConsumeEventsJSON,
    PerpCreateMarketJSON,
    PerpDeactivatePositionJSON,
    PerpLiqBaseOrPositivePnlJSON,
    PerpLiqForceCancelOrdersJSON,
    PerpLiqNegativePnlOrBankruptcyJSON,
    PerpPlaceOrderJSON,
    PerpSettleFeesJSON,
    PerpSettlePnlJSON,
    PerpUpdateFundingJSON,
    Serum3CancelAllOrdersJSON,
    Serum3CancelOrderJSON,
    Serum3CloseOpenOrdersJSON,
    Serum3CreateOpenOrdersJSON,
    Serum3DeregisterMarketJSON,
    Serum3EditMarketJSON,
    Serum3LiqForceCancelOrdersJSON,
    Serum3PlaceOrderJSON,
    Serum3RegisterMarketJSON,
    Serum3SettleFundsJSON,
    StubOracleCloseJSON,
    StubOracleCreateJSON,
    StubOracleSetJSON,
    TokenAddBankJSON,
    TokenDepositJSON,
    TokenDeregisterJSON,
    TokenLiqBankruptcyJSON,
    TokenLiqWithTokenJSON,
    TokenRegisterJSON,
    TokenRegisterTrustlessJSON,
    TokenUpdateIndexAndRateJSON,
    TokenWithdrawJSON,
    AccountBuybackFeesWithMngoJSON,
]


def from_decoded(obj: dict) -> IxGateKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "AccountClose" in obj:
        return AccountClose()
    if "AccountCreate" in obj:
        return AccountCreate()
    if "AccountEdit" in obj:
        return AccountEdit()
    if "AccountExpand" in obj:
        return AccountExpand()
    if "AccountToggleFreeze" in obj:
        return AccountToggleFreeze()
    if "AltExtend" in obj:
        return AltExtend()
    if "AltSet" in obj:
        return AltSet()
    if "FlashLoan" in obj:
        return FlashLoan()
    if "GroupClose" in obj:
        return GroupClose()
    if "GroupCreate" in obj:
        return GroupCreate()
    if "HealthRegion" in obj:
        return HealthRegion()
    if "PerpCancelAllOrders" in obj:
        return PerpCancelAllOrders()
    if "PerpCancelAllOrdersBySide" in obj:
        return PerpCancelAllOrdersBySide()
    if "PerpCancelOrder" in obj:
        return PerpCancelOrder()
    if "PerpCancelOrderByClientOrderId" in obj:
        return PerpCancelOrderByClientOrderId()
    if "PerpCloseMarket" in obj:
        return PerpCloseMarket()
    if "PerpConsumeEvents" in obj:
        return PerpConsumeEvents()
    if "PerpCreateMarket" in obj:
        return PerpCreateMarket()
    if "PerpDeactivatePosition" in obj:
        return PerpDeactivatePosition()
    if "PerpLiqBaseOrPositivePnl" in obj:
        return PerpLiqBaseOrPositivePnl()
    if "PerpLiqForceCancelOrders" in obj:
        return PerpLiqForceCancelOrders()
    if "PerpLiqNegativePnlOrBankruptcy" in obj:
        return PerpLiqNegativePnlOrBankruptcy()
    if "PerpPlaceOrder" in obj:
        return PerpPlaceOrder()
    if "PerpSettleFees" in obj:
        return PerpSettleFees()
    if "PerpSettlePnl" in obj:
        return PerpSettlePnl()
    if "PerpUpdateFunding" in obj:
        return PerpUpdateFunding()
    if "Serum3CancelAllOrders" in obj:
        return Serum3CancelAllOrders()
    if "Serum3CancelOrder" in obj:
        return Serum3CancelOrder()
    if "Serum3CloseOpenOrders" in obj:
        return Serum3CloseOpenOrders()
    if "Serum3CreateOpenOrders" in obj:
        return Serum3CreateOpenOrders()
    if "Serum3DeregisterMarket" in obj:
        return Serum3DeregisterMarket()
    if "Serum3EditMarket" in obj:
        return Serum3EditMarket()
    if "Serum3LiqForceCancelOrders" in obj:
        return Serum3LiqForceCancelOrders()
    if "Serum3PlaceOrder" in obj:
        return Serum3PlaceOrder()
    if "Serum3RegisterMarket" in obj:
        return Serum3RegisterMarket()
    if "Serum3SettleFunds" in obj:
        return Serum3SettleFunds()
    if "StubOracleClose" in obj:
        return StubOracleClose()
    if "StubOracleCreate" in obj:
        return StubOracleCreate()
    if "StubOracleSet" in obj:
        return StubOracleSet()
    if "TokenAddBank" in obj:
        return TokenAddBank()
    if "TokenDeposit" in obj:
        return TokenDeposit()
    if "TokenDeregister" in obj:
        return TokenDeregister()
    if "TokenLiqBankruptcy" in obj:
        return TokenLiqBankruptcy()
    if "TokenLiqWithToken" in obj:
        return TokenLiqWithToken()
    if "TokenRegister" in obj:
        return TokenRegister()
    if "TokenRegisterTrustless" in obj:
        return TokenRegisterTrustless()
    if "TokenUpdateIndexAndRate" in obj:
        return TokenUpdateIndexAndRate()
    if "TokenWithdraw" in obj:
        return TokenWithdraw()
    if "AccountBuybackFeesWithMngo" in obj:
        return AccountBuybackFeesWithMngo()
    raise ValueError("Invalid enum object")


def from_json(obj: IxGateJSON) -> IxGateKind:
    if obj["kind"] == "AccountClose":
        return AccountClose()
    if obj["kind"] == "AccountCreate":
        return AccountCreate()
    if obj["kind"] == "AccountEdit":
        return AccountEdit()
    if obj["kind"] == "AccountExpand":
        return AccountExpand()
    if obj["kind"] == "AccountToggleFreeze":
        return AccountToggleFreeze()
    if obj["kind"] == "AltExtend":
        return AltExtend()
    if obj["kind"] == "AltSet":
        return AltSet()
    if obj["kind"] == "FlashLoan":
        return FlashLoan()
    if obj["kind"] == "GroupClose":
        return GroupClose()
    if obj["kind"] == "GroupCreate":
        return GroupCreate()
    if obj["kind"] == "HealthRegion":
        return HealthRegion()
    if obj["kind"] == "PerpCancelAllOrders":
        return PerpCancelAllOrders()
    if obj["kind"] == "PerpCancelAllOrdersBySide":
        return PerpCancelAllOrdersBySide()
    if obj["kind"] == "PerpCancelOrder":
        return PerpCancelOrder()
    if obj["kind"] == "PerpCancelOrderByClientOrderId":
        return PerpCancelOrderByClientOrderId()
    if obj["kind"] == "PerpCloseMarket":
        return PerpCloseMarket()
    if obj["kind"] == "PerpConsumeEvents":
        return PerpConsumeEvents()
    if obj["kind"] == "PerpCreateMarket":
        return PerpCreateMarket()
    if obj["kind"] == "PerpDeactivatePosition":
        return PerpDeactivatePosition()
    if obj["kind"] == "PerpLiqBaseOrPositivePnl":
        return PerpLiqBaseOrPositivePnl()
    if obj["kind"] == "PerpLiqForceCancelOrders":
        return PerpLiqForceCancelOrders()
    if obj["kind"] == "PerpLiqNegativePnlOrBankruptcy":
        return PerpLiqNegativePnlOrBankruptcy()
    if obj["kind"] == "PerpPlaceOrder":
        return PerpPlaceOrder()
    if obj["kind"] == "PerpSettleFees":
        return PerpSettleFees()
    if obj["kind"] == "PerpSettlePnl":
        return PerpSettlePnl()
    if obj["kind"] == "PerpUpdateFunding":
        return PerpUpdateFunding()
    if obj["kind"] == "Serum3CancelAllOrders":
        return Serum3CancelAllOrders()
    if obj["kind"] == "Serum3CancelOrder":
        return Serum3CancelOrder()
    if obj["kind"] == "Serum3CloseOpenOrders":
        return Serum3CloseOpenOrders()
    if obj["kind"] == "Serum3CreateOpenOrders":
        return Serum3CreateOpenOrders()
    if obj["kind"] == "Serum3DeregisterMarket":
        return Serum3DeregisterMarket()
    if obj["kind"] == "Serum3EditMarket":
        return Serum3EditMarket()
    if obj["kind"] == "Serum3LiqForceCancelOrders":
        return Serum3LiqForceCancelOrders()
    if obj["kind"] == "Serum3PlaceOrder":
        return Serum3PlaceOrder()
    if obj["kind"] == "Serum3RegisterMarket":
        return Serum3RegisterMarket()
    if obj["kind"] == "Serum3SettleFunds":
        return Serum3SettleFunds()
    if obj["kind"] == "StubOracleClose":
        return StubOracleClose()
    if obj["kind"] == "StubOracleCreate":
        return StubOracleCreate()
    if obj["kind"] == "StubOracleSet":
        return StubOracleSet()
    if obj["kind"] == "TokenAddBank":
        return TokenAddBank()
    if obj["kind"] == "TokenDeposit":
        return TokenDeposit()
    if obj["kind"] == "TokenDeregister":
        return TokenDeregister()
    if obj["kind"] == "TokenLiqBankruptcy":
        return TokenLiqBankruptcy()
    if obj["kind"] == "TokenLiqWithToken":
        return TokenLiqWithToken()
    if obj["kind"] == "TokenRegister":
        return TokenRegister()
    if obj["kind"] == "TokenRegisterTrustless":
        return TokenRegisterTrustless()
    if obj["kind"] == "TokenUpdateIndexAndRate":
        return TokenUpdateIndexAndRate()
    if obj["kind"] == "TokenWithdraw":
        return TokenWithdraw()
    if obj["kind"] == "AccountBuybackFeesWithMngo":
        return AccountBuybackFeesWithMngo()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "AccountClose" / borsh.CStruct(),
    "AccountCreate" / borsh.CStruct(),
    "AccountEdit" / borsh.CStruct(),
    "AccountExpand" / borsh.CStruct(),
    "AccountToggleFreeze" / borsh.CStruct(),
    "AltExtend" / borsh.CStruct(),
    "AltSet" / borsh.CStruct(),
    "FlashLoan" / borsh.CStruct(),
    "GroupClose" / borsh.CStruct(),
    "GroupCreate" / borsh.CStruct(),
    "HealthRegion" / borsh.CStruct(),
    "PerpCancelAllOrders" / borsh.CStruct(),
    "PerpCancelAllOrdersBySide" / borsh.CStruct(),
    "PerpCancelOrder" / borsh.CStruct(),
    "PerpCancelOrderByClientOrderId" / borsh.CStruct(),
    "PerpCloseMarket" / borsh.CStruct(),
    "PerpConsumeEvents" / borsh.CStruct(),
    "PerpCreateMarket" / borsh.CStruct(),
    "PerpDeactivatePosition" / borsh.CStruct(),
    "PerpLiqBaseOrPositivePnl" / borsh.CStruct(),
    "PerpLiqForceCancelOrders" / borsh.CStruct(),
    "PerpLiqNegativePnlOrBankruptcy" / borsh.CStruct(),
    "PerpPlaceOrder" / borsh.CStruct(),
    "PerpSettleFees" / borsh.CStruct(),
    "PerpSettlePnl" / borsh.CStruct(),
    "PerpUpdateFunding" / borsh.CStruct(),
    "Serum3CancelAllOrders" / borsh.CStruct(),
    "Serum3CancelOrder" / borsh.CStruct(),
    "Serum3CloseOpenOrders" / borsh.CStruct(),
    "Serum3CreateOpenOrders" / borsh.CStruct(),
    "Serum3DeregisterMarket" / borsh.CStruct(),
    "Serum3EditMarket" / borsh.CStruct(),
    "Serum3LiqForceCancelOrders" / borsh.CStruct(),
    "Serum3PlaceOrder" / borsh.CStruct(),
    "Serum3RegisterMarket" / borsh.CStruct(),
    "Serum3SettleFunds" / borsh.CStruct(),
    "StubOracleClose" / borsh.CStruct(),
    "StubOracleCreate" / borsh.CStruct(),
    "StubOracleSet" / borsh.CStruct(),
    "TokenAddBank" / borsh.CStruct(),
    "TokenDeposit" / borsh.CStruct(),
    "TokenDeregister" / borsh.CStruct(),
    "TokenLiqBankruptcy" / borsh.CStruct(),
    "TokenLiqWithToken" / borsh.CStruct(),
    "TokenRegister" / borsh.CStruct(),
    "TokenRegisterTrustless" / borsh.CStruct(),
    "TokenUpdateIndexAndRate" / borsh.CStruct(),
    "TokenWithdraw" / borsh.CStruct(),
    "AccountBuybackFeesWithMngo" / borsh.CStruct(),
)



import argparse
import asyncio
from base58 import b58decode
from solana.keypair import Keypair

from ..mango_client import MangoClient


async def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--mango-account',
        help='Mango account primary key.',
        required=True
    )

    parser.add_argument(
        '--keypair',
        help='Solana wallet private key, to sign transactions for the Mango account.',
        required=True
    )

    parser.add_argument(
        '--symbol',
        required=True
    )

    parser.add_argument(
        '--side',
        required=True,
        choices=['bids', 'asks']
    )

    parser.add_argument(
        '--price-offset',
        type=float
    )

    parser.add_argument(
        '--peg-limit',
        type=float
    )

    parser.add_argument(
        '--quantity',
        type=float
    )

    args = parser.parse_args()

    mango_client = await MangoClient.connect()

    mango_account = await mango_client.get_mango_account(args.mango_account)

    keypair = Keypair.from_secret_key(b58decode(args.keypair))

    print(
        await mango_client.place_perp_pegged_order(
            mango_account,
            keypair,
            args.symbol,
            args.side,
            args.price_offset,
            args.peg_limit,
            args.quantity
        )
    )

if __name__ == '__main__':
    asyncio.run(main())



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class TokenUpdateIndexAndRateAccounts(typing.TypedDict):
    group: PublicKey
    mint_info: PublicKey
    oracle: PublicKey
    instructions: PublicKey


def token_update_index_and_rate(
    accounts: TokenUpdateIndexAndRateAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["mint_info"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["instructions"], is_signer=False, is_writable=False
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x83\x88\xc2'\x0b2\n\xc6"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class PerpCreateMarketArgs(typing.TypedDict):
    perp_market_index: int
    name: str
    oracle_config: types.oracle_config_params.OracleConfigParams
    base_decimals: int
    quote_lot_size: int
    base_lot_size: int
    maint_base_asset_weight: float
    init_base_asset_weight: float
    maint_base_liab_weight: float
    init_base_liab_weight: float
    maint_overall_asset_weight: float
    init_overall_asset_weight: float
    base_liquidation_fee: float
    maker_fee: float
    taker_fee: float
    min_funding: float
    max_funding: float
    impact_quantity: int
    group_insurance_fund: bool
    fee_penalty: float
    settle_fee_flat: float
    settle_fee_amount_threshold: float
    settle_fee_fraction_low_health: float
    settle_token_index: int
    settle_pnl_limit_factor: float
    settle_pnl_limit_window_size_ts: int
    positive_pnl_liquidation_fee: float


layout = borsh.CStruct(
    "perp_market_index" / borsh.U16,
    "name" / borsh.String,
    "oracle_config" / types.oracle_config_params.OracleConfigParams.layout,
    "base_decimals" / borsh.U8,
    "quote_lot_size" / borsh.I64,
    "base_lot_size" / borsh.I64,
    "maint_base_asset_weight" / borsh.F32,
    "init_base_asset_weight" / borsh.F32,
    "maint_base_liab_weight" / borsh.F32,
    "init_base_liab_weight" / borsh.F32,
    "maint_overall_asset_weight" / borsh.F32,
    "init_overall_asset_weight" / borsh.F32,
    "base_liquidation_fee" / borsh.F32,
    "maker_fee" / borsh.F32,
    "taker_fee" / borsh.F32,
    "min_funding" / borsh.F32,
    "max_funding" / borsh.F32,
    "impact_quantity" / borsh.I64,
    "group_insurance_fund" / borsh.Bool,
    "fee_penalty" / borsh.F32,
    "settle_fee_flat" / borsh.F32,
    "settle_fee_amount_threshold" / borsh.F32,
    "settle_fee_fraction_low_health" / borsh.F32,
    "settle_token_index" / borsh.U16,
    "settle_pnl_limit_factor" / borsh.F32,
    "settle_pnl_limit_window_size_ts" / borsh.U64,
    "positive_pnl_liquidation_fee" / borsh.F32,
)


class PerpCreateMarketAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    oracle: PublicKey
    perp_market: PublicKey
    bids: PublicKey
    asks: PublicKey
    event_queue: PublicKey
    payer: PublicKey


def perp_create_market(
    args: PerpCreateMarketArgs,
    accounts: PerpCreateMarketAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["event_queue"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b']/-\xc3>\xfc "'
    encoded_args = layout.build(
        {
            "perp_market_index": args["perp_market_index"],
            "name": args["name"],
            "oracle_config": args["oracle_config"].to_encodable(),
            "base_decimals": args["base_decimals"],
            "quote_lot_size": args["quote_lot_size"],
            "base_lot_size": args["base_lot_size"],
            "maint_base_asset_weight": args["maint_base_asset_weight"],
            "init_base_asset_weight": args["init_base_asset_weight"],
            "maint_base_liab_weight": args["maint_base_liab_weight"],
            "init_base_liab_weight": args["init_base_liab_weight"],
            "maint_overall_asset_weight": args["maint_overall_asset_weight"],
            "init_overall_asset_weight": args["init_overall_asset_weight"],
            "base_liquidation_fee": args["base_liquidation_fee"],
            "maker_fee": args["maker_fee"],
            "taker_fee": args["taker_fee"],
            "min_funding": args["min_funding"],
            "max_funding": args["max_funding"],
            "impact_quantity": args["impact_quantity"],
            "group_insurance_fund": args["group_insurance_fund"],
            "fee_penalty": args["fee_penalty"],
            "settle_fee_flat": args["settle_fee_flat"],
            "settle_fee_amount_threshold": args["settle_fee_amount_threshold"],
            "settle_fee_fraction_low_health": args["settle_fee_fraction_low_health"],
            "settle_token_index": args["settle_token_index"],
            "settle_pnl_limit_factor": args["settle_pnl_limit_factor"],
            "settle_pnl_limit_window_size_ts": args["settle_pnl_limit_window_size_ts"],
            "positive_pnl_liquidation_fee": args["positive_pnl_liquidation_fee"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class Serum3LiqForceCancelOrdersArgs(typing.TypedDict):
    limit: int


layout = borsh.CStruct("limit" / borsh.U8)


class Serum3LiqForceCancelOrdersAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    open_orders: PublicKey
    serum_market: PublicKey
    serum_program: PublicKey
    serum_market_external: PublicKey
    market_bids: PublicKey
    market_asks: PublicKey
    market_event_queue: PublicKey
    market_base_vault: PublicKey
    market_quote_vault: PublicKey
    market_vault_signer: PublicKey
    quote_bank: PublicKey
    quote_vault: PublicKey
    base_bank: PublicKey
    base_vault: PublicKey


def serum3_liq_force_cancel_orders(
    args: Serum3LiqForceCancelOrdersArgs,
    accounts: Serum3LiqForceCancelOrdersAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["open_orders"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["serum_market"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_market_external"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["market_bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["market_asks"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["market_event_queue"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["market_base_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["market_quote_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["market_vault_signer"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["quote_bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["quote_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["base_bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["base_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x1f\xaa_]X6\t\xe7"
    encoded_args = layout.build(
        {
            "limit": args["limit"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



import decimal
import sys

from mango_explorer_v4.accounts.bank import Bank
from mango_explorer_v4.types.i80f48 import I80F48


class BankHelper():
    @staticmethod
    def native_deposits(bank: Bank) -> decimal.Decimal:
        return bank.indexed_deposits.to_decimal() * bank.deposit_index.to_decimal()

    @staticmethod
    def native_borrows(bank: Bank) -> decimal.Decimal:
        return bank.indexed_borrows.to_decimal() * bank.borrow_index.to_decimal()

    @staticmethod
    def scaled_init_asset_weight(bank: Bank, price: decimal.Decimal) -> I80F48:
        deposits_quote = BankHelper.native_deposits(bank) * price

        if bank.deposit_weight_scale_start_quote >= sys.maxsize or deposits_quote <= bank.deposit_weight_scale_start_quote:
            return bank.init_asset_weight

        return I80F48.from_decimal(bank.init_asset_weight.to_decimal() * decimal.Decimal(bank.deposit_weight_scale_start_quote) / deposits_quote)

    @staticmethod
    def scaled_init_liab_weight(bank: Bank, price: decimal.Decimal) -> I80F48:
        borrows_quote = BankHelper.native_borrows(bank) * price

        if bank.borrow_weight_scale_start_quote >= sys.maxsize or borrows_quote <= bank.borrow_weight_scale_start_quote:
            return bank.init_liab_weight

        return I80F48.from_decimal(bank.init_liab_weight.to_decimal() * borrows_quote / decimal.Decimal(bank.borrow_weight_scale_start_quote))






from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class Serum3DeregisterMarketAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    serum_market: PublicKey
    index_reservation: PublicKey
    sol_destination: PublicKey


def serum3_deregister_market(
    accounts: Serum3DeregisterMarketAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["serum_market"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["index_reservation"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["sol_destination"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x11\xa4*\xde\x97\xa0\x18\xb5"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



import asyncio
from ..mango_client import MangoClient


async def main():
    mango_client = await MangoClient.connect()

    mango_accounts = await mango_client.get_all_mango_accounts()

    for mango_account, balances in zip(
        mango_accounts,
        await asyncio.gather(*[mango_client.balances(mango_account) for mango_account in mango_accounts])
    ):
        print(mango_account.public_key, [{**entry, 'balance': round(entry['balance'], 2)} for entry in balances])

if __name__ == '__main__':
    asyncio.run(main())


from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class PerpPlaceOrderPeggedArgs(typing.TypedDict):
    side: types.side.SideKind
    price_offset_lots: int
    peg_limit: int
    max_base_lots: int
    max_quote_lots: int
    client_order_id: int
    order_type: types.place_order_type.PlaceOrderTypeKind
    reduce_only: bool
    expiry_timestamp: int
    limit: int
    max_oracle_staleness_slots: int


layout = borsh.CStruct(
    "side" / types.side.layout,
    "price_offset_lots" / borsh.I64,
    "peg_limit" / borsh.I64,
    "max_base_lots" / borsh.I64,
    "max_quote_lots" / borsh.I64,
    "client_order_id" / borsh.U64,
    "order_type" / types.place_order_type.layout,
    "reduce_only" / borsh.Bool,
    "expiry_timestamp" / borsh.U64,
    "limit" / borsh.U8,
    "max_oracle_staleness_slots" / borsh.I32,
)


class PerpPlaceOrderPeggedAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    perp_market: PublicKey
    bids: PublicKey
    asks: PublicKey
    event_queue: PublicKey
    oracle: PublicKey


def perp_place_order_pegged(
    args: PerpPlaceOrderPeggedArgs,
    accounts: PerpPlaceOrderPeggedAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["event_queue"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xc0<\x99\xa2\xf6\xc82W"
    encoded_args = layout.build(
        {
            "side": args["side"].to_encodable(),
            "price_offset_lots": args["price_offset_lots"],
            "peg_limit": args["peg_limit"],
            "max_base_lots": args["max_base_lots"],
            "max_quote_lots": args["max_quote_lots"],
            "client_order_id": args["client_order_id"],
            "order_type": args["order_type"].to_encodable(),
            "reduce_only": args["reduce_only"],
            "expiry_timestamp": args["expiry_timestamp"],
            "limit": args["limit"],
            "max_oracle_staleness_slots": args["max_oracle_staleness_slots"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class AltSetArgs(typing.TypedDict):
    index: int


layout = borsh.CStruct("index" / borsh.U8)


class AltSetAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    address_lookup_table: PublicKey


def alt_set(
    args: AltSetArgs,
    accounts: AltSetAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["address_lookup_table"], is_signer=False, is_writable=True
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xebD\x91 ;i7\x19"
    encoded_args = layout.build(
        {
            "index": args["index"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class PerpCancelOrderByClientOrderIdArgs(typing.TypedDict):
    client_order_id: int


layout = borsh.CStruct("client_order_id" / borsh.U64)


class PerpCancelOrderByClientOrderIdAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    perp_market: PublicKey
    bids: PublicKey
    asks: PublicKey


def perp_cancel_order_by_client_order_id(
    args: PerpCancelOrderByClientOrderIdArgs,
    accounts: PerpCancelOrderByClientOrderIdAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"J\xfa8O\xce\xad\xa3f"
    encoded_args = layout.build(
        {
            "client_order_id": args["client_order_id"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
from . import (
    token_equity,
    perp_equity,
)
import typing
from dataclasses import dataclass
from construct import Container, Construct
import borsh_construct as borsh


class EquityJSON(typing.TypedDict):
    tokens: list[token_equity.TokenEquityJSON]
    perps: list[perp_equity.PerpEquityJSON]


@dataclass
class Equity:
    layout: typing.ClassVar = borsh.CStruct(
        "tokens" / borsh.Vec(typing.cast(Construct, token_equity.TokenEquity.layout)),
        "perps" / borsh.Vec(typing.cast(Construct, perp_equity.PerpEquity.layout)),
    )
    tokens: list[token_equity.TokenEquity]
    perps: list[perp_equity.PerpEquity]

    @classmethod
    def from_decoded(cls, obj: Container) -> "Equity":
        return cls(
            tokens=list(
                map(
                    lambda item: token_equity.TokenEquity.from_decoded(item), obj.tokens
                )
            ),
            perps=list(
                map(lambda item: perp_equity.PerpEquity.from_decoded(item), obj.perps)
            ),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "tokens": list(map(lambda item: item.to_encodable(), self.tokens)),
            "perps": list(map(lambda item: item.to_encodable(), self.perps)),
        }

    def to_json(self) -> EquityJSON:
        return {
            "tokens": list(map(lambda item: item.to_json(), self.tokens)),
            "perps": list(map(lambda item: item.to_json(), self.perps)),
        }

    @classmethod
    def from_json(cls, obj: EquityJSON) -> "Equity":
        return cls(
            tokens=list(
                map(
                    lambda item: token_equity.TokenEquity.from_json(item), obj["tokens"]
                )
            ),
            perps=list(
                map(lambda item: perp_equity.PerpEquity.from_json(item), obj["perps"])
            ),
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class FixedJSON(typing.TypedDict):
    kind: typing.Literal["Fixed"]


class OraclePeggedJSON(typing.TypedDict):
    kind: typing.Literal["OraclePegged"]


@dataclass
class Fixed:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Fixed"

    @classmethod
    def to_json(cls) -> FixedJSON:
        return FixedJSON(
            kind="Fixed",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Fixed": {},
        }


@dataclass
class OraclePegged:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "OraclePegged"

    @classmethod
    def to_json(cls) -> OraclePeggedJSON:
        return OraclePeggedJSON(
            kind="OraclePegged",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "OraclePegged": {},
        }


BookSideOrderTreeKind = typing.Union[Fixed, OraclePegged]
BookSideOrderTreeJSON = typing.Union[FixedJSON, OraclePeggedJSON]


def from_decoded(obj: dict) -> BookSideOrderTreeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Fixed" in obj:
        return Fixed()
    if "OraclePegged" in obj:
        return OraclePegged()
    raise ValueError("Invalid enum object")


def from_json(obj: BookSideOrderTreeJSON) -> BookSideOrderTreeKind:
    if obj["kind"] == "Fixed":
        return Fixed()
    if obj["kind"] == "OraclePegged":
        return OraclePegged()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Fixed" / borsh.CStruct(), "OraclePegged" / borsh.CStruct())



from mango_explorer_v4.types.perp_position import PerpPosition
from mango_explorer_v4.accounts.perp_market import PerpMarket
from mango_explorer_v4.helpers.perp_market import PerpMarketHelper
from decimal import Decimal
from dataclasses import dataclass


@dataclass
class PerpPositionHelper():
    @staticmethod
    def is_active(perp_position: PerpPosition):
        return perp_position.market_index != 65535

    @staticmethod
    def equity(perp_position: PerpPosition, perp_market: PerpMarket, perp_market_oracle_price: Decimal):
        if perp_position.market_index != perp_market.perp_market_index:
            raise ValueError(f"Perp position does not belong to perp market")

        base_lots_to_quote_currency_converter = Decimal(perp_market.base_lot_size) * perp_market_oracle_price * Decimal(10 ** (6 - perp_market.base_decimals))

        base_lots = perp_position.base_position_lots + perp_position.taker_base_lots

        taker_quote = perp_position.taker_quote_lots * perp_market.quote_lot_size

        unsettled_funding = PerpPositionHelper.unsettled_funding(perp_position, perp_market)

        quote_current = perp_position.quote_position_native.to_decimal() - unsettled_funding + taker_quote

        # print(perp_market.perp_market_index, perp_market_oracle_price * Decimal(10 ** (6 - perp_market.base_decimals)), base_lots_to_quote_currency_converter, base_lots, unsettled_funding, taker_quote, quote_current)

        return (base_lots * base_lots_to_quote_currency_converter + quote_current) / Decimal(1e6)

    @staticmethod
    def unsettled_funding(perp_position: PerpPosition, perp_market: PerpMarket) -> Decimal:
        if perp_position.market_index != perp_market.perp_market_index:
            raise ValueError(f"Perp position does not belong to perp market")

        if perp_position.base_position_lots == 0:
            return Decimal(0)

        minuend, subtrahend = (
            [perp_market.long_funding, perp_position.long_settled_funding]
            if perp_position.base_position_lots > 0 else
            [perp_market.short_funding, perp_position.short_settled_funding]
        )

        multiplier = perp_position.base_position_lots

        return (minuend.to_decimal() - subtrahend.to_decimal()) * multiplier

    @staticmethod
    def unsettled_pnl(
        perp_position: PerpPosition, perp_market: PerpMarket, oracle_price: Decimal
    ) -> Decimal:
        if perp_position.market_index != perp_market.perp_market_index:
            raise ValueError(f"Perp position does not belong to perp market")

        return (
            perp_position.quote_position_native.to_decimal()
            + Decimal(PerpPositionHelper.base_position_native(perp_position, perp_market))
            * Decimal(oracle_price) * Decimal(10 ** (6 - perp_market.base_decimals))
            # ^ 6 is the number of decimals for the insurance mint TODO: change this when that does
        ) / Decimal(1e6)

    @staticmethod
    def base_position_native(
        perp_position: PerpPosition, perp_market: PerpMarket
    ) -> int:
        return perp_position.base_position_lots * perp_market.base_lot_size

    @staticmethod
    def base_position_ui(perp_position: PerpPosition, perp_market: PerpMarket, use_event_queue: bool = False):
        if perp_position.market_index != perp_market.perp_market_index:
            raise ValueError("Perp position doesn't belong to perp market")

        return PerpMarketHelper.base_lots_to_ui(
            perp_market,
            perp_position.base_position_lots + perp_position.taker_base_lots
            if use_event_queue
            else perp_position.base_position_lots
        )

    @staticmethod
    def has_open_orders(perp_position: PerpPosition):
        return any([
            value != 0
            for value in [
                perp_position.asks_base_lots,
                perp_position.bids_base_lots,
                perp_position.taker_base_lots,
                perp_position.taker_quote_lots
            ]
        ])

    @staticmethod
    def has_open_fills(perp_position: PerpPosition):
        return perp_position.taker_base_lots != 0 or perp_position.taker_quote_lots != 0






import argparse
import asyncio

from base58 import b58decode
from solana.keypair import Keypair
from solana.transaction import Transaction

from ..mango_client import MangoClient


async def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--mango-account',
        help='Mango account primary key.',
        required=True
    )

    parser.add_argument(
        '--keypair',
        help='Solana wallet private key, to sign transactions for the Mango account.',
        required=True
    )

    args = parser.parse_args()

    mango_client = await MangoClient.connect()

    mango_account = await mango_client.get_mango_account(args.mango_account)

    keypair = Keypair.from_secret_key(b58decode(args.keypair))

    symbol = 'SOL/USDC'

    orderbook = await mango_client.orderbook_l2(symbol)

    mid_price = (orderbook['bids'][0][0] + orderbook['asks'][0][0]) / 2

    spread = 500 / 1e4  # bps

    orders = [{
        'symbol': symbol,
        'side': 'bids',
        'price': mid_price - (mid_price * spread),
        'size': 1,  # TODO: Have here the minimum order size
    }, {
        'symbol': symbol,
        'side': 'asks',
        'price': mid_price + (mid_price * spread),
        'size': 1,
    }]

    serum3_cancel_all_orders_ix = mango_client.make_serum3_cancel_all_orders_ix(mango_account, symbol)

    serum3_place_order_ixs = map(lambda order: mango_client.make_serum3_place_order_ix(mango_account, **order), orders)

    tx = Transaction()

    tx.add(serum3_cancel_all_orders_ix, *serum3_place_order_ixs)

    recent_blockhash = str((await mango_client.connection.get_latest_blockhash()).value.blockhash)

    try:
        await mango_client.connection.send_transaction(tx, keypair, recent_blockhash=recent_blockhash); print(f"Quoted: {orders}")
    except Exception as exception:
        print(f"{exception}")

if __name__ == '__main__':
    asyncio.run(main())



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solana.publickey import PublicKey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class OutEventJSON(typing.TypedDict):
    event_type: int
    side: int
    owner_slot: int
    padding0: list[int]
    timestamp: int
    seq_num: int
    owner: str
    quantity: int
    padding1: list[int]


@dataclass
class OutEvent:
    layout: typing.ClassVar = borsh.CStruct(
        "event_type" / borsh.U8,
        "side" / borsh.U8,
        "owner_slot" / borsh.U8,
        "padding0" / borsh.U8[5],
        "timestamp" / borsh.U64,
        "seq_num" / borsh.U64,
        "owner" / BorshPubkey,
        "quantity" / borsh.I64,
        "padding1" / borsh.U8[144],
    )
    event_type: int
    side: int
    owner_slot: int
    padding0: list[int]
    timestamp: int
    seq_num: int
    owner: PublicKey
    quantity: int
    padding1: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "OutEvent":
        return cls(
            event_type=obj.event_type,
            side=obj.side,
            owner_slot=obj.owner_slot,
            padding0=obj.padding0,
            timestamp=obj.timestamp,
            seq_num=obj.seq_num,
            owner=obj.owner,
            quantity=obj.quantity,
            padding1=obj.padding1,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "event_type": self.event_type,
            "side": self.side,
            "owner_slot": self.owner_slot,
            "padding0": self.padding0,
            "timestamp": self.timestamp,
            "seq_num": self.seq_num,
            "owner": self.owner,
            "quantity": self.quantity,
            "padding1": self.padding1,
        }

    def to_json(self) -> OutEventJSON:
        return {
            "event_type": self.event_type,
            "side": self.side,
            "owner_slot": self.owner_slot,
            "padding0": self.padding0,
            "timestamp": self.timestamp,
            "seq_num": self.seq_num,
            "owner": str(self.owner),
            "quantity": self.quantity,
            "padding1": self.padding1,
        }

    @classmethod
    def from_json(cls, obj: OutEventJSON) -> "OutEvent":
        return cls(
            event_type=obj["event_type"],
            side=obj["side"],
            owner_slot=obj["owner_slot"],
            padding0=obj["padding0"],
            timestamp=obj["timestamp"],
            seq_num=obj["seq_num"],
            owner=PublicKey(obj["owner"]),
            quantity=obj["quantity"],
            padding1=obj["padding1"],
        )



from __future__ import annotations
from . import (
    any_node,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class OrderTreeNodesJSON(typing.TypedDict):
    order_tree_type: int
    padding: list[int]
    bump_index: int
    free_list_len: int
    free_list_head: int
    reserved: list[int]
    nodes: list[any_node.AnyNodeJSON]


@dataclass
class OrderTreeNodes:
    layout: typing.ClassVar = borsh.CStruct(
        "order_tree_type" / borsh.U8,
        "padding" / borsh.U8[3],
        "bump_index" / borsh.U32,
        "free_list_len" / borsh.U32,
        "free_list_head" / borsh.U32,
        "reserved" / borsh.U8[512],
        "nodes" / any_node.AnyNode.layout[1024],
    )
    order_tree_type: int
    padding: list[int]
    bump_index: int
    free_list_len: int
    free_list_head: int
    reserved: list[int]
    nodes: list[any_node.AnyNode]

    @classmethod
    def from_decoded(cls, obj: Container) -> "OrderTreeNodes":
        return cls(
            order_tree_type=obj.order_tree_type,
            padding=obj.padding,
            bump_index=obj.bump_index,
            free_list_len=obj.free_list_len,
            free_list_head=obj.free_list_head,
            reserved=obj.reserved,
            nodes=list(
                map(lambda item: any_node.AnyNode.from_decoded(item), obj.nodes)
            ),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "order_tree_type": self.order_tree_type,
            "padding": self.padding,
            "bump_index": self.bump_index,
            "free_list_len": self.free_list_len,
            "free_list_head": self.free_list_head,
            "reserved": self.reserved,
            "nodes": list(map(lambda item: item.to_encodable(), self.nodes)),
        }

    def to_json(self) -> OrderTreeNodesJSON:
        return {
            "order_tree_type": self.order_tree_type,
            "padding": self.padding,
            "bump_index": self.bump_index,
            "free_list_len": self.free_list_len,
            "free_list_head": self.free_list_head,
            "reserved": self.reserved,
            "nodes": list(map(lambda item: item.to_json(), self.nodes)),
        }

    @classmethod
    def from_json(cls, obj: OrderTreeNodesJSON) -> "OrderTreeNodes":
        return cls(
            order_tree_type=obj["order_tree_type"],
            padding=obj["padding"],
            bump_index=obj["bump_index"],
            free_list_len=obj["free_list_len"],
            free_list_head=obj["free_list_head"],
            reserved=obj["reserved"],
            nodes=list(
                map(lambda item: any_node.AnyNode.from_json(item), obj["nodes"])
            ),
        )



from solana.publickey import PublicKey

SERUM_PROGRAM_ID = PublicKey('srmqPvymJeFKQ4zGQed1GFppgkRHL9kaELCbyksJtPX')

RUST_U64_MAX = 18446744073709551615

RUST_I64_MAX = 9223372036854775807

RUST_I64_MIN = -9223372036854775807

QUOTE_DECIMALS = 6



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class UnknownJSON(typing.TypedDict):
    kind: typing.Literal["Unknown"]


class LiqTokenBankruptcyJSON(typing.TypedDict):
    kind: typing.Literal["LiqTokenBankruptcy"]


class LiqTokenWithTokenJSON(typing.TypedDict):
    kind: typing.Literal["LiqTokenWithToken"]


class Serum3LiqForceCancelOrdersJSON(typing.TypedDict):
    kind: typing.Literal["Serum3LiqForceCancelOrders"]


class Serum3PlaceOrderJSON(typing.TypedDict):
    kind: typing.Literal["Serum3PlaceOrder"]


class Serum3SettleFundsJSON(typing.TypedDict):
    kind: typing.Literal["Serum3SettleFunds"]


class TokenWithdrawJSON(typing.TypedDict):
    kind: typing.Literal["TokenWithdraw"]


@dataclass
class Unknown:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Unknown"

    @classmethod
    def to_json(cls) -> UnknownJSON:
        return UnknownJSON(
            kind="Unknown",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Unknown": {},
        }


@dataclass
class LiqTokenBankruptcy:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "LiqTokenBankruptcy"

    @classmethod
    def to_json(cls) -> LiqTokenBankruptcyJSON:
        return LiqTokenBankruptcyJSON(
            kind="LiqTokenBankruptcy",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "LiqTokenBankruptcy": {},
        }


@dataclass
class LiqTokenWithToken:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "LiqTokenWithToken"

    @classmethod
    def to_json(cls) -> LiqTokenWithTokenJSON:
        return LiqTokenWithTokenJSON(
            kind="LiqTokenWithToken",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "LiqTokenWithToken": {},
        }


@dataclass
class Serum3LiqForceCancelOrders:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "Serum3LiqForceCancelOrders"

    @classmethod
    def to_json(cls) -> Serum3LiqForceCancelOrdersJSON:
        return Serum3LiqForceCancelOrdersJSON(
            kind="Serum3LiqForceCancelOrders",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3LiqForceCancelOrders": {},
        }


@dataclass
class Serum3PlaceOrder:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "Serum3PlaceOrder"

    @classmethod
    def to_json(cls) -> Serum3PlaceOrderJSON:
        return Serum3PlaceOrderJSON(
            kind="Serum3PlaceOrder",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3PlaceOrder": {},
        }


@dataclass
class Serum3SettleFunds:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "Serum3SettleFunds"

    @classmethod
    def to_json(cls) -> Serum3SettleFundsJSON:
        return Serum3SettleFundsJSON(
            kind="Serum3SettleFunds",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Serum3SettleFunds": {},
        }


@dataclass
class TokenWithdraw:
    discriminator: typing.ClassVar = 6
    kind: typing.ClassVar = "TokenWithdraw"

    @classmethod
    def to_json(cls) -> TokenWithdrawJSON:
        return TokenWithdrawJSON(
            kind="TokenWithdraw",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "TokenWithdraw": {},
        }


LoanOriginationFeeInstructionKind = typing.Union[
    Unknown,
    LiqTokenBankruptcy,
    LiqTokenWithToken,
    Serum3LiqForceCancelOrders,
    Serum3PlaceOrder,
    Serum3SettleFunds,
    TokenWithdraw,
]
LoanOriginationFeeInstructionJSON = typing.Union[
    UnknownJSON,
    LiqTokenBankruptcyJSON,
    LiqTokenWithTokenJSON,
    Serum3LiqForceCancelOrdersJSON,
    Serum3PlaceOrderJSON,
    Serum3SettleFundsJSON,
    TokenWithdrawJSON,
]


def from_decoded(obj: dict) -> LoanOriginationFeeInstructionKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Unknown" in obj:
        return Unknown()
    if "LiqTokenBankruptcy" in obj:
        return LiqTokenBankruptcy()
    if "LiqTokenWithToken" in obj:
        return LiqTokenWithToken()
    if "Serum3LiqForceCancelOrders" in obj:
        return Serum3LiqForceCancelOrders()
    if "Serum3PlaceOrder" in obj:
        return Serum3PlaceOrder()
    if "Serum3SettleFunds" in obj:
        return Serum3SettleFunds()
    if "TokenWithdraw" in obj:
        return TokenWithdraw()
    raise ValueError("Invalid enum object")


def from_json(
    obj: LoanOriginationFeeInstructionJSON,
) -> LoanOriginationFeeInstructionKind:
    if obj["kind"] == "Unknown":
        return Unknown()
    if obj["kind"] == "LiqTokenBankruptcy":
        return LiqTokenBankruptcy()
    if obj["kind"] == "LiqTokenWithToken":
        return LiqTokenWithToken()
    if obj["kind"] == "Serum3LiqForceCancelOrders":
        return Serum3LiqForceCancelOrders()
    if obj["kind"] == "Serum3PlaceOrder":
        return Serum3PlaceOrder()
    if obj["kind"] == "Serum3SettleFunds":
        return Serum3SettleFunds()
    if obj["kind"] == "TokenWithdraw":
        return TokenWithdraw()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Unknown" / borsh.CStruct(),
    "LiqTokenBankruptcy" / borsh.CStruct(),
    "LiqTokenWithToken" / borsh.CStruct(),
    "Serum3LiqForceCancelOrders" / borsh.CStruct(),
    "Serum3PlaceOrder" / borsh.CStruct(),
    "Serum3SettleFunds" / borsh.CStruct(),
    "TokenWithdraw" / borsh.CStruct(),
)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.sysvar import SYSVAR_RENT_PUBKEY
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class TokenRegisterTrustlessArgs(typing.TypedDict):
    token_index: int
    name: str


layout = borsh.CStruct("token_index" / borsh.U16, "name" / borsh.String)


class TokenRegisterTrustlessAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    mint: PublicKey
    bank: PublicKey
    vault: PublicKey
    mint_info: PublicKey
    oracle: PublicKey
    payer: PublicKey


def token_register_trustless(
    args: TokenRegisterTrustlessArgs,
    accounts: TokenRegisterTrustlessAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mint_info"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYSVAR_RENT_PUBKEY, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"8-#\r\xfd\xfe:P"
    encoded_args = layout.build(
        {
            "token_index": args["token_index"],
            "name": args["name"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
from . import (
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class Serum3InfoJSON(typing.TypedDict):
    reserved_base: i80f48.I80F48JSON
    reserved_quote: i80f48.I80F48JSON
    base_index: int
    quote_index: int
    market_index: int
    has_zero_funds: bool


@dataclass
class Serum3Info:
    layout: typing.ClassVar = borsh.CStruct(
        "reserved_base" / i80f48.I80F48.layout,
        "reserved_quote" / i80f48.I80F48.layout,
        "base_index" / borsh.U64,
        "quote_index" / borsh.U64,
        "market_index" / borsh.U16,
        "has_zero_funds" / borsh.Bool,
    )
    reserved_base: i80f48.I80F48
    reserved_quote: i80f48.I80F48
    base_index: int
    quote_index: int
    market_index: int
    has_zero_funds: bool

    @classmethod
    def from_decoded(cls, obj: Container) -> "Serum3Info":
        return cls(
            reserved_base=i80f48.I80F48.from_decoded(obj.reserved_base),
            reserved_quote=i80f48.I80F48.from_decoded(obj.reserved_quote),
            base_index=obj.base_index,
            quote_index=obj.quote_index,
            market_index=obj.market_index,
            has_zero_funds=obj.has_zero_funds,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "reserved_base": self.reserved_base.to_encodable(),
            "reserved_quote": self.reserved_quote.to_encodable(),
            "base_index": self.base_index,
            "quote_index": self.quote_index,
            "market_index": self.market_index,
            "has_zero_funds": self.has_zero_funds,
        }

    def to_json(self) -> Serum3InfoJSON:
        return {
            "reserved_base": self.reserved_base.to_json(),
            "reserved_quote": self.reserved_quote.to_json(),
            "base_index": self.base_index,
            "quote_index": self.quote_index,
            "market_index": self.market_index,
            "has_zero_funds": self.has_zero_funds,
        }

    @classmethod
    def from_json(cls, obj: Serum3InfoJSON) -> "Serum3Info":
        return cls(
            reserved_base=i80f48.I80F48.from_json(obj["reserved_base"]),
            reserved_quote=i80f48.I80F48.from_json(obj["reserved_quote"]),
            base_index=obj["base_index"],
            quote_index=obj["quote_index"],
            market_index=obj["market_index"],
            has_zero_funds=obj["has_zero_funds"],
        )



import typing
from dataclasses import dataclass
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class Serum3MarketJSON(typing.TypedDict):
    group: str
    base_token_index: int
    quote_token_index: int
    reduce_only: int
    padding1: list[int]
    name: list[int]
    serum_program: str
    serum_market_external: str
    market_index: int
    bump: int
    padding2: list[int]
    registration_time: int
    reserved: list[int]


@dataclass
class Serum3Market:
    discriminator: typing.ClassVar = b"u\x07\xb6\xf6`h\x88\x84"
    layout: typing.ClassVar = borsh.CStruct(
        "group" / BorshPubkey,
        "base_token_index" / borsh.U16,
        "quote_token_index" / borsh.U16,
        "reduce_only" / borsh.U8,
        "padding1" / borsh.U8[3],
        "name" / borsh.U8[16],
        "serum_program" / BorshPubkey,
        "serum_market_external" / BorshPubkey,
        "market_index" / borsh.U16,
        "bump" / borsh.U8,
        "padding2" / borsh.U8[5],
        "registration_time" / borsh.U64,
        "reserved" / borsh.U8[128],
    )
    group: PublicKey
    base_token_index: int
    quote_token_index: int
    reduce_only: int
    padding1: list[int]
    name: list[int]
    serum_program: PublicKey
    serum_market_external: PublicKey
    market_index: int
    bump: int
    padding2: list[int]
    registration_time: int
    reserved: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.Optional["Serum3Market"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Serum3Market"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Serum3Market"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Serum3Market":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Serum3Market.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            group=dec.group,
            base_token_index=dec.base_token_index,
            quote_token_index=dec.quote_token_index,
            reduce_only=dec.reduce_only,
            padding1=dec.padding1,
            name=dec.name,
            serum_program=dec.serum_program,
            serum_market_external=dec.serum_market_external,
            market_index=dec.market_index,
            bump=dec.bump,
            padding2=dec.padding2,
            registration_time=dec.registration_time,
            reserved=dec.reserved,
        )

    def to_json(self) -> Serum3MarketJSON:
        return {
            "group": str(self.group),
            "base_token_index": self.base_token_index,
            "quote_token_index": self.quote_token_index,
            "reduce_only": self.reduce_only,
            "padding1": self.padding1,
            "name": self.name,
            "serum_program": str(self.serum_program),
            "serum_market_external": str(self.serum_market_external),
            "market_index": self.market_index,
            "bump": self.bump,
            "padding2": self.padding2,
            "registration_time": self.registration_time,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: Serum3MarketJSON) -> "Serum3Market":
        return cls(
            group=PublicKey(obj["group"]),
            base_token_index=obj["base_token_index"],
            quote_token_index=obj["quote_token_index"],
            reduce_only=obj["reduce_only"],
            padding1=obj["padding1"],
            name=obj["name"],
            serum_program=PublicKey(obj["serum_program"]),
            serum_market_external=PublicKey(obj["serum_market_external"]),
            market_index=obj["market_index"],
            bump=obj["bump"],
            padding2=obj["padding2"],
            registration_time=obj["registration_time"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
from . import (
    prices,
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class TokenInfoJSON(typing.TypedDict):
    token_index: int
    maint_asset_weight: i80f48.I80F48JSON
    init_asset_weight: i80f48.I80F48JSON
    init_scaled_asset_weight: i80f48.I80F48JSON
    maint_liab_weight: i80f48.I80F48JSON
    init_liab_weight: i80f48.I80F48JSON
    init_scaled_liab_weight: i80f48.I80F48JSON
    prices: prices.PricesJSON
    balance_native: i80f48.I80F48JSON


@dataclass
class TokenInfo:
    layout: typing.ClassVar = borsh.CStruct(
        "token_index" / borsh.U16,
        "maint_asset_weight" / i80f48.I80F48.layout,
        "init_asset_weight" / i80f48.I80F48.layout,
        "init_scaled_asset_weight" / i80f48.I80F48.layout,
        "maint_liab_weight" / i80f48.I80F48.layout,
        "init_liab_weight" / i80f48.I80F48.layout,
        "init_scaled_liab_weight" / i80f48.I80F48.layout,
        "prices" / prices.Prices.layout,
        "balance_native" / i80f48.I80F48.layout,
    )
    token_index: int
    maint_asset_weight: i80f48.I80F48
    init_asset_weight: i80f48.I80F48
    init_scaled_asset_weight: i80f48.I80F48
    maint_liab_weight: i80f48.I80F48
    init_liab_weight: i80f48.I80F48
    init_scaled_liab_weight: i80f48.I80F48
    prices: prices.Prices
    balance_native: i80f48.I80F48

    @classmethod
    def from_decoded(cls, obj: Container) -> "TokenInfo":
        return cls(
            token_index=obj.token_index,
            maint_asset_weight=i80f48.I80F48.from_decoded(obj.maint_asset_weight),
            init_asset_weight=i80f48.I80F48.from_decoded(obj.init_asset_weight),
            init_scaled_asset_weight=i80f48.I80F48.from_decoded(
                obj.init_scaled_asset_weight
            ),
            maint_liab_weight=i80f48.I80F48.from_decoded(obj.maint_liab_weight),
            init_liab_weight=i80f48.I80F48.from_decoded(obj.init_liab_weight),
            init_scaled_liab_weight=i80f48.I80F48.from_decoded(
                obj.init_scaled_liab_weight
            ),
            prices=prices.Prices.from_decoded(obj.prices),
            balance_native=i80f48.I80F48.from_decoded(obj.balance_native),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "token_index": self.token_index,
            "maint_asset_weight": self.maint_asset_weight.to_encodable(),
            "init_asset_weight": self.init_asset_weight.to_encodable(),
            "init_scaled_asset_weight": self.init_scaled_asset_weight.to_encodable(),
            "maint_liab_weight": self.maint_liab_weight.to_encodable(),
            "init_liab_weight": self.init_liab_weight.to_encodable(),
            "init_scaled_liab_weight": self.init_scaled_liab_weight.to_encodable(),
            "prices": self.prices.to_encodable(),
            "balance_native": self.balance_native.to_encodable(),
        }

    def to_json(self) -> TokenInfoJSON:
        return {
            "token_index": self.token_index,
            "maint_asset_weight": self.maint_asset_weight.to_json(),
            "init_asset_weight": self.init_asset_weight.to_json(),
            "init_scaled_asset_weight": self.init_scaled_asset_weight.to_json(),
            "maint_liab_weight": self.maint_liab_weight.to_json(),
            "init_liab_weight": self.init_liab_weight.to_json(),
            "init_scaled_liab_weight": self.init_scaled_liab_weight.to_json(),
            "prices": self.prices.to_json(),
            "balance_native": self.balance_native.to_json(),
        }

    @classmethod
    def from_json(cls, obj: TokenInfoJSON) -> "TokenInfo":
        return cls(
            token_index=obj["token_index"],
            maint_asset_weight=i80f48.I80F48.from_json(obj["maint_asset_weight"]),
            init_asset_weight=i80f48.I80F48.from_json(obj["init_asset_weight"]),
            init_scaled_asset_weight=i80f48.I80F48.from_json(
                obj["init_scaled_asset_weight"]
            ),
            maint_liab_weight=i80f48.I80F48.from_json(obj["maint_liab_weight"]),
            init_liab_weight=i80f48.I80F48.from_json(obj["init_liab_weight"]),
            init_scaled_liab_weight=i80f48.I80F48.from_json(
                obj["init_scaled_liab_weight"]
            ),
            prices=prices.Prices.from_json(obj["prices"]),
            balance_native=i80f48.I80F48.from_json(obj["balance_native"]),
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.sysvar import SYSVAR_RENT_PUBKEY
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import PROGRAM_ID


class Serum3CreateOpenOrdersAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    serum_market: PublicKey
    serum_program: PublicKey
    serum_market_external: PublicKey
    open_orders: PublicKey
    payer: PublicKey


def serum3_create_open_orders(
    accounts: Serum3CreateOpenOrdersAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["serum_market"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_market_external"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["open_orders"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYSVAR_RENT_PUBKEY, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x04\xf4#(US\xcb\xfa"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
from . import (
    i80f48,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class TokenEquityJSON(typing.TypedDict):
    token_index: int
    value: i80f48.I80F48JSON


@dataclass
class TokenEquity:
    layout: typing.ClassVar = borsh.CStruct(
        "token_index" / borsh.U16, "value" / i80f48.I80F48.layout
    )
    token_index: int
    value: i80f48.I80F48

    @classmethod
    def from_decoded(cls, obj: Container) -> "TokenEquity":
        return cls(
            token_index=obj.token_index, value=i80f48.I80F48.from_decoded(obj.value)
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"token_index": self.token_index, "value": self.value.to_encodable()}

    def to_json(self) -> TokenEquityJSON:
        return {"token_index": self.token_index, "value": self.value.to_json()}

    @classmethod
    def from_json(cls, obj: TokenEquityJSON) -> "TokenEquity":
        return cls(
            token_index=obj["token_index"], value=i80f48.I80F48.from_json(obj["value"])
        )



from mango_explorer_v4.types.i80f48 import I80F48
from decimal import Decimal

# TODO: This can be appended to the cla
def parse(i80f48: I80F48) -> float:
    return float(Decimal(i80f48.val) / Decimal(2 ** 48))



from mango_explorer_v4.types.perp_info import PerpInfo
from mango_explorer_v4.types.health_type import HealthTypeKind, Init, LiquidationEnd
from mango_explorer_v4.types.i80f48 import I80F48
from mango_explorer_v4.helpers.prices import PricesHelper
from decimal import Decimal

class PerpInfoHelper:
    @staticmethod
    def unweighted_health_contribution(perp_info: PerpInfo, health_type: HealthTypeKind) -> I80F48:
        def order_execution_case(perp_info: PerpInfo, orders_base_lots: Decimal, order_price: I80F48) -> I80F48:
            net_base_native = perp_info.base_lots + orders_base_lots * perp_info.base_lot_size

            if health_type in [Init(), LiquidationEnd()]:
                if net_base_native < 0:
                    weight = perp_info.init_base_liab_weight
                else:
                    weight = perp_info.init_base_asset_weight
            else:
                if net_base_native < 0:
                    weight = perp_info.maint_base_liab_weight
                else:
                    weight = perp_info.maint_base_asset_weight

            if net_base_native < 0:
                base_price = PricesHelper.liab(perp_info.prices, health_type)
            else:
                base_price = PricesHelper.asset(perp_info.prices, health_type)

            # Total value of the order-execution adjusted base position
            base_health = net_base_native * weight.to_decimal() * base_price

            orders_base_native = orders_base_lots * perp_info.base_lot_size

            order_quote = orders_base_native * -1 * order_price.to_decimal()

            return I80F48.from_decimal(base_health + order_quote)

        bids_case = order_execution_case(
            perp_info,
            Decimal(perp_info.bids_base_lots),
            I80F48.from_decimal(PricesHelper.liab(perp_info.prices, health_type))
        ).to_decimal()

        asks_case = order_execution_case(
            perp_info,
            Decimal(perp_info.bids_base_lots) * -1,
            I80F48.from_decimal(PricesHelper.asset(perp_info.prices, health_type))
        ).to_decimal()

        worst_case = min(bids_case, asks_case)

        return I80F48.from_decimal(perp_info.quote.to_decimal() + worst_case)

    @staticmethod
    def health_contribution(perp_info: PerpInfo, health_type: HealthTypeKind) -> I80F48:
        contrib = PerpInfoHelper.unweighted_health_contribution(perp_info, health_type).to_decimal()

        if contrib > 0:
            asset_weight = (perp_info.init_overall_asset_weight if health_type in [Init(), LiquidationEnd()] else perp_info.maint_overall_asset_weight).to_decimal()

            return I80F48.from_decimal(asset_weight * contrib)

        return I80F48.from_decimal(contrib)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class AccountToggleFreezeArgs(typing.TypedDict):
    freeze: bool


layout = borsh.CStruct("freeze" / borsh.Bool)


class AccountToggleFreezeAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    admin: PublicKey


def account_toggle_freeze(
    args: AccountToggleFreezeArgs,
    accounts: AccountToggleFreezeAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x0b\xf3\xe9\xf7WO`\xc6"
    encoded_args = layout.build(
        {
            "freeze": args["freeze"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class Serum3MarketIndexJSON(typing.TypedDict):
    val: int


@dataclass
class Serum3MarketIndex:
    layout: typing.ClassVar = borsh.CStruct("val" / borsh.U16)
    val: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "Serum3MarketIndex":
        return cls(val=obj.val)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"val": self.val}

    def to_json(self) -> Serum3MarketIndexJSON:
        return {"val": self.val}

    @classmethod
    def from_json(cls, obj: Serum3MarketIndexJSON) -> "Serum3MarketIndex":
        return cls(val=obj["val"])



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class UnknownJSON(typing.TypedDict):
    kind: typing.Literal["Unknown"]


class SwapJSON(typing.TypedDict):
    kind: typing.Literal["Swap"]


@dataclass
class Unknown:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Unknown"

    @classmethod
    def to_json(cls) -> UnknownJSON:
        return UnknownJSON(
            kind="Unknown",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Unknown": {},
        }


@dataclass
class Swap:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Swap"

    @classmethod
    def to_json(cls) -> SwapJSON:
        return SwapJSON(
            kind="Swap",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Swap": {},
        }


FlashLoanTypeKind = typing.Union[Unknown, Swap]
FlashLoanTypeJSON = typing.Union[UnknownJSON, SwapJSON]


def from_decoded(obj: dict) -> FlashLoanTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Unknown" in obj:
        return Unknown()
    if "Swap" in obj:
        return Swap()
    raise ValueError("Invalid enum object")


def from_json(obj: FlashLoanTypeJSON) -> FlashLoanTypeKind:
    if obj["kind"] == "Unknown":
        return Unknown()
    if obj["kind"] == "Swap":
        return Swap()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Unknown" / borsh.CStruct(), "Swap" / borsh.CStruct())



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class FlashLoanEndArgs(typing.TypedDict):
    flash_loan_type: types.flash_loan_type.FlashLoanTypeKind


layout = borsh.CStruct("flash_loan_type" / types.flash_loan_type.layout)


class FlashLoanEndAccounts(typing.TypedDict):
    account: PublicKey
    owner: PublicKey


def flash_loan_end(
    args: FlashLoanEndArgs,
    accounts: FlashLoanEndAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xb2\xaa\x02N\xf0\x17\xbe\xb2"
    encoded_args = layout.build(
        {
            "flash_loan_type": args["flash_loan_type"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class Serum3SettleFundsV2Args(typing.TypedDict):
    fees_to_dao: bool


layout = borsh.CStruct("fees_to_dao" / borsh.Bool)


class Serum3SettleFundsV2Accounts(typing.TypedDict):
    v1: V1Nested
    v2: V2Nested


class V1Nested(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    open_orders: PublicKey
    serum_market: PublicKey
    serum_program: PublicKey
    serum_market_external: PublicKey
    market_base_vault: PublicKey
    market_quote_vault: PublicKey
    market_vault_signer: PublicKey
    quote_bank: PublicKey
    quote_vault: PublicKey
    base_bank: PublicKey
    base_vault: PublicKey


class V2Nested(typing.TypedDict):
    quote_oracle: PublicKey
    base_oracle: PublicKey


class V1Nested(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    open_orders: PublicKey
    serum_market: PublicKey
    serum_program: PublicKey
    serum_market_external: PublicKey
    market_base_vault: PublicKey
    market_quote_vault: PublicKey
    market_vault_signer: PublicKey
    quote_bank: PublicKey
    quote_vault: PublicKey
    base_bank: PublicKey
    base_vault: PublicKey


def serum3_settle_funds_v2(
    args: Serum3SettleFundsV2Args,
    accounts: Serum3SettleFundsV2Accounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["v1"]["group"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["v1"]["account"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["v1"]["owner"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["v1"]["open_orders"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["v1"]["serum_market"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["v1"]["serum_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["v1"]["serum_market_external"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["v1"]["market_base_vault"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["v1"]["market_quote_vault"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["v1"]["market_vault_signer"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["v1"]["quote_bank"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["v1"]["quote_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["v1"]["base_bank"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["v1"]["base_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["v2"]["quote_oracle"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["v2"]["base_oracle"], is_signer=False, is_writable=False
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x05\xddk\x17\xf7\xd3/\xdd"
    encoded_args = layout.build(
        {
            "fees_to_dao": args["fees_to_dao"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from .bank import Bank, BankJSON
from .group import Group, GroupJSON
from .mango_account import MangoAccount, MangoAccountJSON
from .mint_info import MintInfo, MintInfoJSON
from .stub_oracle import StubOracle, StubOracleJSON
from .book_side import BookSide, BookSideJSON
from .event_queue import EventQueue, EventQueueJSON
from .perp_market import PerpMarket, PerpMarketJSON
from .serum3_market import Serum3Market, Serum3MarketJSON
from .serum3_market_index_reservation import (
    Serum3MarketIndexReservation,
    Serum3MarketIndexReservationJSON,
)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class AccountBuybackFeesWithMngoArgs(typing.TypedDict):
    max_buyback: int


layout = borsh.CStruct("max_buyback" / borsh.U64)


class AccountBuybackFeesWithMngoAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    dao_account: PublicKey
    mngo_bank: PublicKey
    mngo_oracle: PublicKey
    fees_bank: PublicKey
    fees_oracle: PublicKey


def account_buyback_fees_with_mngo(
    args: AccountBuybackFeesWithMngoArgs,
    accounts: AccountBuybackFeesWithMngoAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["dao_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mngo_bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mngo_oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["fees_bank"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["fees_oracle"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x0c\x19B\x8b\x16`f\xc1"
    encoded_args = layout.build(
        {
            "max_buyback": args["max_buyback"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.system_program import SYS_PROGRAM_ID
from solana.sysvar import SYSVAR_RENT_PUBKEY
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class GroupCreateArgs(typing.TypedDict):
    group_num: int
    testing: int
    version: int


layout = borsh.CStruct(
    "group_num" / borsh.U32, "testing" / borsh.U8, "version" / borsh.U8
)


class GroupCreateAccounts(typing.TypedDict):
    group: PublicKey
    creator: PublicKey
    insurance_mint: PublicKey
    insurance_vault: PublicKey
    payer: PublicKey


def group_create(
    args: GroupCreateArgs,
    accounts: GroupCreateAccounts,
    program_id: PublicKey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["creator"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["insurance_mint"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["insurance_vault"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYSVAR_RENT_PUBKEY, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xe2R\xefwk\x88\xa6\xf0"
    encoded_args = layout.build(
        {
            "group_num": args["group_num"],
            "testing": args["testing"],
            "version": args["version"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class InnerNodeJSON(typing.TypedDict):
    tag: int
    padding: list[int]
    prefix_len: int
    key: int
    children: list[int]
    child_earliest_expiry: list[int]
    reserved: list[int]


@dataclass
class InnerNode:
    layout: typing.ClassVar = borsh.CStruct(
        "tag" / borsh.U8,
        "padding" / borsh.U8[3],
        "prefix_len" / borsh.U32,
        "key" / borsh.U128,
        "children" / borsh.U32[2],
        "child_earliest_expiry" / borsh.U64[2],
        "reserved" / borsh.U8[72],
    )
    tag: int
    padding: list[int]
    prefix_len: int
    key: int
    children: list[int]
    child_earliest_expiry: list[int]
    reserved: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "InnerNode":
        return cls(
            tag=obj.tag,
            padding=obj.padding,
            prefix_len=obj.prefix_len,
            key=obj.key,
            children=obj.children,
            child_earliest_expiry=obj.child_earliest_expiry,
            reserved=obj.reserved,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "tag": self.tag,
            "padding": self.padding,
            "prefix_len": self.prefix_len,
            "key": self.key,
            "children": self.children,
            "child_earliest_expiry": self.child_earliest_expiry,
            "reserved": self.reserved,
        }

    def to_json(self) -> InnerNodeJSON:
        return {
            "tag": self.tag,
            "padding": self.padding,
            "prefix_len": self.prefix_len,
            "key": self.key,
            "children": self.children,
            "child_earliest_expiry": self.child_earliest_expiry,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(cls, obj: InnerNodeJSON) -> "InnerNode":
        return cls(
            tag=obj["tag"],
            padding=obj["padding"],
            prefix_len=obj["prefix_len"],
            key=obj["key"],
            children=obj["children"],
            child_earliest_expiry=obj["child_earliest_expiry"],
            reserved=obj["reserved"],
        )
